"""
Apply ADMM on a single image from DiffuserCam dataset: https://github.com/Waller-Lab/LenslessLearning

Or download a subset here: https://drive.switch.ch/index.php/s/vmAZzryGI8U8rcE

```
python scripts/recon/admm_mirflickr.py
```

"""

import hydra
from hydra.utils import to_absolute_path
import glob
import numpy as np
from lensless.utils.image import print_image_info
from lensless.utils.io import load_psf
from lensless.utils.plot import plot_image
import matplotlib.pyplot as plt
from datetime import datetime
import pathlib as plib
import os
import random

# import click
import time
from lensless.utils.io import load_image
from lensless.recon.mirflickr import ADMM_MIRFLICKR, postprocess
from lensless.eval.metric import mse, psnr, ssim, lpips
import matplotlib

font = {"family": "DejaVu Sans", "size": 18}
matplotlib.rc("font", **font)


@hydra.main(
    version_base=None, config_path="../../configs", config_name="apply_admm_single_mirflickr"
)
def apply_admm(config):
    data = to_absolute_path(config.dataset)
    fid = config.fid
    gamma = config.display.gamma
    single_psf = config.preprocess.single_psf
    disp = config.display.disp
    save = config.save
    admm_param = config.admm

    if fid is None:
        # random pick file
        fn = glob.glob(os.path.join(os.path.join(data, "diffuser"), "*.npy"))
        fn = [_fn.split("im")[1] for _fn in fn]
        fn = [int(_fn.split(".")[0]) for _fn in fn]
        fid = random.choice(fn)
        print(f"Reconstruction : {fid}")
    downsample = 4  # has to be this for collected data!

    # form file path
    psf_fp = os.path.join(data, "psf.tiff")
    dataset_dir = os.path.join(data, "dataset")
    if os.path.isdir(dataset_dir):
        lensless_fp = os.path.join(dataset_dir, f"diffuser_images/im{fid}.npy")
        lensed_fp = os.path.join(dataset_dir, f"ground_truth_lensed/im{fid}.npy")
    else:
        lensless_fp = os.path.join(data, f"diffuser/im{fid}.npy")
        lensed_fp = os.path.join(data, f"lensed/im{fid}.npy")

    # initialize plot
    _, ax = plt.subplots(ncols=2, nrows=2, figsize=(15, 10))

    # diffuser data
    print("Diffuser data")
    diffuser = np.load(lensless_fp)
    print_image_info(diffuser)
    _ax = plot_image(postprocess(diffuser), gamma=gamma, ax=ax[0, 0])
    _ax.set_title("Lensless")

    # Lensed data
    print("\nLensed data")
    lensed = np.load(lensed_fp)
    print_image_info(lensed)
    lensed = postprocess(lensed)
    _ax = plot_image(lensed, gamma=gamma, ax=ax[0, 1])
    _ax.set_title("Lensed")

    # PSF
    print("\nPSF data")
    psf = load_image(fp=psf_fp, verbose=True)
    _ax = plot_image(psf, gamma=2.2, ax=ax[1, 0])
    _ax.set_title("PSF (gamma corrected)")

    # RECONSTRUCTION
    # -- prepare data
    print("\nPrepared PSF data")
    psf_float, background = load_psf(
        psf_fp,
        downsample=downsample,
        return_float=True,
        return_bg=True,
        bg_pix=(0, 15),
        single_psf=single_psf,
    )
    print_image_info(psf_float)

    print("\nPrepared diffuser data")
    diffuser_prep = diffuser - background
    diffuser_prep = np.clip(diffuser_prep, a_min=0, a_max=1)
    diffuser_prep /= np.linalg.norm(diffuser_prep.ravel())
    print_image_info(diffuser_prep)

    # apply ADMM
    if save:
        save = os.path.basename(lensless_fp).split(".")[0]
        timestamp = datetime.now().strftime("_%d%m%d%Y_%Hh%M")
        save = "admm_" + save + timestamp
        save = plib.Path(__file__).parent / save
        save.mkdir(exist_ok=False)
    recon = ADMM_MIRFLICKR(psf_float, **admm_param)
    recon.set_data(diffuser_prep)
    start_time = time.time()
    recon.apply(n_iter=admm_param["n_iter"], disp_iter=disp, save=save, gamma=gamma, ax=ax[1, 1])
    proc_time = time.time() - start_time
    print(f"Processing time : {proc_time} seconds")
    ax[1, 1].set_title(f"ADMM ({admm_param.n_iter} iterations)")

    # -- compute metrics
    print("\nReconstruction")
    est = recon.get_image_estimate()
    print_image_info(est)

    print("\nMSE", mse(lensed, est))
    print("PSNR", psnr(lensed, est))
    print("SSIM", ssim(lensed, est))
    print("LPIPS", lpips(lensed, est))

    plt.show()


if __name__ == "__main__":
    apply_admm()



"""
Check maximum pixel value of images and check for saturation / underexposure.

```
python scripts/measure/analyze_measured_dataset.py folder=PATH
```
"""

import hydra
from hydra.utils import to_absolute_path
import glob
import os
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
import time
import tqdm


@hydra.main(version_base=None, config_path="../../configs", config_name="analyze_dataset")
def analyze_dataset(config):

    folder = config.dataset_path
    desired_range = config.desired_range
    delete_bad = config.delete_bad
    start_idx = config.start_idx

    assert (
        folder is not None
    ), "Must specify folder to analyze in config or through command line (folder=PATH)."

    # get all PNG files in folder
    files = sorted(glob.glob(os.path.join(folder, "*.png")))
    print("Found {} files".format(len(files)))
    if start_idx is not None:
        files = files[start_idx:]
        print("Starting at file {}".format(files[0]))
    if config.n_files is not None:
        files = files[: config.n_files]
        print("Analyzing first {} files".format(len(files)))

    # loop over files for maximum value
    max_vals = []
    n_bad_files = 0
    bad_files = []
    start_time = time.time()
    for fn in tqdm.tqdm(files):
        im = np.array(Image.open(fn))
        max_val = im.max()
        max_vals.append(max_val)

        # if out of desired range, print filename
        if max_val < desired_range[0] or max_val > desired_range[1]:
            # print("File {} has max value {}".format(fn, max_val))
            n_bad_files += 1
            bad_files.append(fn)

            if delete_bad:
                os.remove(fn)
                print("REMOVED file {}".format(fn))
            else:
                print("File {} has max value {}".format(fn, max_val))

    proc_time = time.time() - start_time
    print("Went through {} files in {:.2f} seconds".format(len(files), proc_time))
    print(
        "Found {} / {} bad files  ({}%)".format(
            n_bad_files, len(files), n_bad_files / len(files) * 100
        )
    )

    # command line input on whether to delete bad files
    if not delete_bad:
        response = None
        while response not in ["yes", "no"]:
            response = input("Delete bad files: [yes|no] : ")
        if response == "yes":
            for _fn in bad_files:
                os.remove(_fn)
        else:
            print("Not deleting bad files")

    # plot histogram
    output_folder = os.getcwd()
    output_fp = os.path.join(output_folder, "max_vals.png")
    plt.hist(max_vals, bins=100)
    plt.savefig(output_fp)

    print("Saved histogram to {}".format(output_fp))


if __name__ == "__main__":
    analyze_dataset()



from lensless.utils.io import load_psf, load_image, save_image
from lensless import ADMM
import numpy as np
from lensless.utils.simulation import FarFieldSimulator
import torch
import matplotlib.pyplot as plt


# lensless_fp = "000001.png"
lensless_fp = "/scratch/bezzam/celeba/celeba_adafruit_random_30cm_2mm_20231004_26K/000001.png"
original_fp = "/scratch/bezzam/celeba/img_align_celeba/000001.jpg"
psf_fp = "rpi_hq_adafruit_psf_2mm/raw_data_rgb.png"
downsample = 1
flip = True
scene2mask = 0.25
mask2sensor = 0.002
object_height = 0.33
sensor = "rpi_hq"
torch_device = "cuda"
vertical_shift = -117
# vertical_shift = 0
# horizontal_shift = -80
horizontal_shift = -25

crop = {
    "vertical": [30, 560],
    "horizontal": [285, 720],
}


psf, background = load_psf(
    psf_fp,
    downsample=downsample * 4,  # PSF is 4x the resolution of the images
    return_float=True,
    return_bg=True,
    flip=flip,
    bg_pix=(0, 15),
)
psf = torch.from_numpy(psf).type(torch.float32).to(torch_device)


print(f"PSF shape : {psf.shape}")
print(f"PSF min : {psf.min()}")
print(f"PSF max : {psf.max()}")
print(f"PSF dtype : {psf.dtype}")
print(f"PSF norm : {psf.norm()}")


lensless = load_image(lensless_fp, downsample=downsample, flip=flip)
original = load_image(original_fp)


# convert to torch
lensless = torch.from_numpy(lensless)
original = torch.from_numpy(original)


simulator = FarFieldSimulator(
    is_torch=True,
    output_dim=tuple(psf.shape[-3:-1]),
    scene2mask=scene2mask,
    mask2sensor=mask2sensor,
    object_height=object_height,
    sensor=sensor,
)
lensed = simulator.propagate_image(original, return_object_plane=True)
if vertical_shift is not None:
    lensed = torch.roll(lensed, vertical_shift, dims=-3)
if horizontal_shift is not None:
    lensed = torch.roll(lensed, horizontal_shift, dims=-2)


print(f"Lensless shape : {lensless.shape}")
print(f"Lensed shape : {lensed.shape}")


recon = ADMM(psf)
recon.set_data(lensless.to(torch_device))
res = recon.apply(disp_iter=None, plot=False, n_iter=10)
res_np = res[0].cpu().numpy()
res_np = res_np / res_np.max()
save_image(res_np, f"lensless_recon.png")
lensed_np = lensed.cpu().numpy()
save_image(lensed_np, f"lensed.png")

plt.figure()
plt.imshow(lensed_np, alpha=0.4)
plt.imshow(res_np, alpha=0.7)
plt.savefig("overlay_lensed_recon.png")



# #############################################################################
# slm.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


import os
import numpy as np
from lensless.hardware.utils import check_username_hostname
from lensless.utils.io import get_ctypes
from slm_controller.hardware import SLMParam, slm_devices
from waveprop.spherical import spherical_prop
from waveprop.color import ColorSystem
from waveprop.rs import angular_spectrum
from waveprop.slm import get_centers
from waveprop.devices import SLMParam as SLMParam_wp
from scipy.ndimage import rotate as rotate_func


try:
    import torch
    from torchvision import transforms

    torch_available = True
except ImportError:
    torch_available = False


SUPPORTED_DEVICE = {
    "adafruit": "~/slm-controller/examples/adafruit_slm.py",
    "nokia": "~/slm-controller/examples/nokia_slm.py",
    "holoeye": "~/slm-controller/examples/holoeye_slm.py",
}


def set_programmable_mask(pattern, device, rpi_username, rpi_hostname, verbose=False):
    """
    Set LCD pattern on Raspberry Pi.

    This function assumes that `slm-controller <https://github.com/ebezzam/slm-controller>`_
    is installed on the Raspberry Pi.

    Parameters
    ----------
    pattern : :py:class:`~numpy.ndarray`
        Pattern to set on programmable mask.
    device : str
        Name of device to set pattern on. Supported devices: "adafruit", "nokia", "holoeye".
    rpi_username : str
        Username of Raspberry Pi.
    rpi_hostname : str
        Hostname of Raspberry Pi.

    """

    client = check_username_hostname(rpi_username, rpi_hostname)

    # get path to python executable on Raspberry Pi
    rpi_python = "~/slm-controller/slm_controller_env/bin/python"
    assert (
        device in SUPPORTED_DEVICE.keys()
    ), f"Device {device} not supported. Supported devices: {SUPPORTED_DEVICE.keys()}"
    script = SUPPORTED_DEVICE[device]

    # check that pattern is correct shape
    expected_shape = slm_devices[device][SLMParam.SLM_SHAPE]
    if not slm_devices[device][SLMParam.MONOCHROME]:
        expected_shape = (3, *expected_shape)
    assert (
        pattern.shape == expected_shape
    ), f"Pattern shape {pattern.shape} does not match expected shape {expected_shape}"

    # save pattern
    pattern_fn = "tmp_pattern.npy"
    local_path = os.path.join(os.getcwd(), pattern_fn)
    np.save(local_path, pattern)

    # copy pattern to Raspberry Pi
    remote_path = f"~/{pattern_fn}"
    if verbose:
        print(f"PUTTING {local_path} to {remote_path}")

    os.system(
        'scp %s "%s@%s:%s" >/dev/null 2>&1' % (local_path, rpi_username, rpi_hostname, remote_path)
    )
    # # -- not sure why this doesn't work... permission denied
    # sftp = client.open_sftp()
    # sftp.put(local_path, remote_path, confirm=True)
    # sftp.close()

    # run script on Raspberry Pi to set mask pattern
    command = f"{rpi_python} {script} --file_path {remote_path}"
    if verbose:
        print(f"COMMAND : {command}")
    _stdin, _stdout, _stderr = client.exec_command(command)
    if verbose:
        print(_stdout.read().decode())
    client.close()

    os.remove(local_path)


def get_programmable_mask(
    vals,
    sensor,
    slm_param,
    rotate=None,
    flipud=False,
    nbits=8,
    color_filter=None,
):
    """
    Get mask as a numpy or torch array. Return same type.

    Parameters
    ----------
    vals : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
        Values to set on programmable mask.
    sensor : :py:class:`~lensless.hardware.sensor.VirtualSensor`
        Sensor object.
    slm_param : dict
        SLM parameters.
    rotate : float, optional
        Rotation angle in degrees.
    flipud : bool, optional
        Flip mask vertically.
    nbits : int, optional
        Number of bits/levels to quantize mask to.

    """

    use_torch = False
    if torch_available:
        use_torch = isinstance(vals, torch.Tensor)
    dtype = vals.dtype

    # -- prepare SLM mask
    n_active_slm_pixels = vals.shape
    n_color_filter = np.prod(slm_param["color_filter"].shape[:2])
    pixel_pitch = slm_param[SLMParam_wp.PITCH]
    centers = get_centers(n_active_slm_pixels, pixel_pitch=pixel_pitch)

    if color_filter is None and SLMParam_wp.COLOR_FILTER in slm_param.keys():
        color_filter = slm_param[SLMParam_wp.COLOR_FILTER]
        if isinstance(vals, torch.Tensor):
            color_filter = torch.tensor(color_filter).to(vals)

    if color_filter is not None:

        if isinstance(color_filter, np.ndarray):
            if flipud:
                color_filter = np.flipud(color_filter)
        elif isinstance(color_filter, torch.Tensor):
            if flipud:
                color_filter = torch.flip(color_filter, dims=(0,))
        else:
            raise ValueError("color_filter must be numpy array or torch tensor")

    d1 = sensor.pitch
    _height_pixel, _width_pixel = (slm_param[SLMParam_wp.CELL_SIZE] / d1).astype(int)

    if use_torch:
        mask = torch.zeros((n_color_filter,) + tuple(sensor.resolution)).to(vals)
        slm_vals_flat = vals.flatten()
    else:
        mask = np.zeros((n_color_filter,) + tuple(sensor.resolution), dtype=dtype)
        slm_vals_flat = vals.reshape(-1)

    for i, _center in enumerate(centers):

        _center_pixel = (_center / d1 + sensor.resolution / 2).astype(int)
        _center_top_left_pixel = (
            _center_pixel[0] - np.floor(_height_pixel / 2).astype(int),
            _center_pixel[1] + 1 - np.floor(_width_pixel / 2).astype(int),
        )

        color_filter_idx = i // n_active_slm_pixels[1] % n_color_filter

        mask_val = slm_vals_flat[i] * color_filter[color_filter_idx][0]
        if isinstance(mask_val, np.ndarray):
            mask_val = mask_val[:, np.newaxis, np.newaxis]
        elif isinstance(mask_val, torch.Tensor):
            mask_val = mask_val.unsqueeze(-1).unsqueeze(-1)
        mask[
            :,
            _center_top_left_pixel[0] : _center_top_left_pixel[0] + _height_pixel,
            _center_top_left_pixel[1] : _center_top_left_pixel[1] + _width_pixel,
        ] = mask_val

    # # quantize mask
    # if use_torch:
    #     mask = mask / torch.max(mask)
    #     mask = torch.round(mask * (2**nbits - 1)) / (2**nbits - 1)
    # else:
    #     mask = mask / np.max(mask)
    #     mask = np.round(mask * (2**nbits - 1)) / (2**nbits - 1)

    # rotate
    if rotate is not None:
        if use_torch:
            mask = transforms.functional.rotate(mask, angle=rotate)
        else:
            mask = rotate_func(mask, axes=(2, 1), angle=rotate, reshape=False)

    return mask


def adafruit_sub2full(
    subpattern,
    center,
):
    sub_shape = subpattern.shape
    controllable_shape = (3, sub_shape[0] // 3, sub_shape[1])
    subpattern_rgb = subpattern.reshape(controllable_shape, order="F")
    subpattern_rgb *= 255

    # pad to full pattern
    pattern = np.zeros((3, 128, 160), dtype=np.uint8)
    topleft = [center[0] - controllable_shape[1] // 2, center[1] - controllable_shape[2] // 2]
    pattern[
        :,
        topleft[0] : topleft[0] + controllable_shape[1],
        topleft[1] : topleft[1] + controllable_shape[2],
    ] = subpattern_rgb.astype(np.uint8)
    return pattern


def full2subpattern(
    pattern,
    shape,
    center,
    slm=None,
):
    shape = np.array(shape)
    center = np.array(center)

    # extract region
    idx_1 = center[0] - shape[0] // 2
    idx_2 = center[1] - shape[1] // 2
    subpattern = pattern[:, idx_1 : idx_1 + shape[0], idx_2 : idx_2 + shape[1]]
    subpattern = subpattern / 255.0
    if slm == "adafruit":
        # flatten color channel along rows
        subpattern = subpattern.reshape((-1, subpattern.shape[-1]), order="F")
    return subpattern


def get_intensity_psf(
    mask,
    waveprop=False,
    sensor=None,
    scene2mask=None,
    mask2sensor=None,
    color_system=None,
):
    """
    Get intensity PSF from mask pattern. Return same type of data.

    Parameters
    ----------
    mask : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
        Mask pattern.
    waveprop : bool, optional
        Whether to use wave propagation to compute PSF. Default is False,
        namely to return squared intensity of mask pattern as the PSF (i.e.,
        no wave propagation and just shadow of pattern).
    sensor : :py:class:`~lensless.hardware.sensor.VirtualSensor`
        Sensor object. Not used if ``waveprop=False``.
    scene2mask : float
        Distance from scene to mask. Not used if ``waveprop=False``.
    mask2sensor : float
        Distance from mask to sensor. Not used if ``waveprop=False``.
    color_system : :py:class:`~waveprop.color.ColorSystem`, optional
        Color system. Not used if ``waveprop=False``.

    """
    if color_system is None:
        color_system = ColorSystem.rgb()

    is_torch = False
    device = None
    if torch_available and isinstance(mask, torch.Tensor):
        is_torch = True
        device = mask.device

    dtype = mask.dtype
    ctype, _ = get_ctypes(dtype, is_torch)

    if is_torch:
        psfs = torch.zeros(mask.shape, dtype=ctype, device=device)
    else:
        psfs = np.zeros(mask.shape, dtype=ctype)

    if waveprop:

        assert sensor is not None, "sensor must be specified"
        assert scene2mask is not None, "scene2mask must be specified"
        assert mask2sensor is not None, "mask2sensor must be specified"

        assert (
            len(color_system.wv) == mask.shape[0]
        ), "Number of wavelengths must match number of color channels"

        # spherical wavefronts to mask
        spherical_wavefront = spherical_prop(
            in_shape=sensor.resolution,
            d1=sensor.pitch,
            wv=color_system.wv,
            dz=scene2mask,
            return_psf=True,
            is_torch=is_torch,
            device=device,
            dtype=dtype,
        )
        u_in = spherical_wavefront * mask

        # free space propagation to sensor
        for i, wv in enumerate(color_system.wv):
            psfs[i], _, _ = angular_spectrum(
                u_in=u_in[i],
                wv=wv,
                d1=sensor.pitch,
                dz=mask2sensor,
                dtype=dtype,
                device=device,
            )

    else:

        psfs = mask

    # -- intensity PSF
    if is_torch:
        psf_in = torch.square(torch.abs(psfs))
    else:
        psf_in = np.square(np.abs(psfs))

    return psf_in



# #############################################################################
# trainable_recon.py
# ==================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# #############################################################################

import pathlib as plib
from matplotlib import pyplot as plt
from lensless.recon.recon import ReconstructionAlgorithm
from lensless.utils.plot import plot_image

try:
    import torch

except ImportError:
    raise ImportError("Pytorch is require to use trainable reconstruction algorithms.")


class TrainableReconstructionAlgorithm(ReconstructionAlgorithm, torch.nn.Module):
    """
    Abstract class for defining lensless imaging reconstruction algorithms with trainable parameters.

    The following abstract methods need to be defined:

    * ``_update``: updating state variables at each iterations.
    * ``reset``: reset state variables.
    * ``_form_image``: any pre-processing that needs to be done in order to view the image estimate, e.g. reshaping or clipping.

    One advantage of deriving from this abstract class is that functionality for
    iterating, saving, and visualization is already implemented, namely in the
    ``apply`` method.

    Consequently, using a reconstruction algorithm that derives from it boils down
    to four steps:

    1. Creating an instance of the reconstruction algorithm.
    2. Training the algorithm
    3. Setting the data.
    4. Applying the algorithm.


    """

    def __init__(
        self,
        psf,
        dtype=None,
        n_iter=1,
        pre_process=None,
        post_process=None,
        skip_unrolled=False,
        return_unrolled_output=False,
        legacy_denoiser=False,
        **kwargs,
    ):
        """
        Base constructor. Derived constructor may define new state variables
        here and also reset them in `reset`.

        Parameters
        ----------

            psf : :py:class:`~torch.Tensor`
                Point spread function (PSF) that models forward propagation.
                Must be of shape (depth, height, width, channels) even if
                depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
                to load a PSF from a file such that it is in the correct format.
            dtype : float32 or float64
                Data type to use for optimization.
            n_iter : int
                Number of iterations for unrolled algorithm.
            pre_process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
                If :py:class:`function` : Function to apply to the image estimate before algorithm. Its input most be (image to process, noise_level), where noise_level is a learnable parameter. If it include aditional learnable parameters, they will not be added to the parameter list of the algorithm. To allow for traning, the function must be autograd compatible.
                If :py:class:`~torch.nn.Module` : A DruNet compatible network to apply to the image estimate before algorithm. See ``utils.image.apply_denoiser`` for more details. The network will be included as a submodule of the algorithm and its parameters will be added to the parameter list of the algorithm. If this isn't intended behavior, set requires_grad=False.
            post_process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
                If :py:class:`function` : Function to apply to the image estimate after the whole algorithm. Its input most be (image to process, noise_level), where noise_level is a learnable parameter. If it include aditional learnable parameters, they will not be added to the parameter list of the algorithm. To allow for traning, the function must be autograd compatible.
                If :py:class:`~torch.nn.Module` : A DruNet compatible network to apply to the image estimate after the whole algorithm. See ``utils.image.apply_denoiser`` for more details. The network will be included as a submodule of the algorithm and its parameters will be added to the parameter list of the algorithm. If this isn't intended behavior, set requires_grad=False.
            skip_unrolled : bool, optional
                Whether to skip the unrolled algorithm and only apply the pre- or post-processor block (e.g. to just use a U-Net for reconstruction).
            return_unrolled_output : bool, optional
                Whether to return the output of the unrolled algorithm if also using a post-processor block.
        """
        assert isinstance(psf, torch.Tensor), "PSF must be a torch.Tensor"
        super(TrainableReconstructionAlgorithm, self).__init__(
            psf, dtype=dtype, n_iter=n_iter, **kwargs
        )

        self._legacy_denoiser = legacy_denoiser
        self.set_pre_process(pre_process)
        self.set_post_process(post_process)
        self.skip_unrolled = skip_unrolled
        self.return_unrolled_output = return_unrolled_output
        if self.return_unrolled_output:
            assert (
                post_process is not None
            ), "If return_unrolled_output is True, post_process must be defined."
        if self.skip_unrolled:
            assert (
                post_process is not None or pre_process is not None
            ), "If skip_unrolled is True, pre_process or post_process must be defined."

    def _prepare_process_block(self, process):
        """
        Method for preparing the pre or post process block.
        Parameters
        ----------
            process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
                Pre or post process block to prepare.
        """
        if isinstance(process, torch.nn.Module):
            # If the post_process is a torch module, we assume it is a DruNet like network.
            from lensless.recon.utils import get_drunet_function, get_drunet_function_v2

            process_model = process
            if self._legacy_denoiser:
                process_function = get_drunet_function(
                    process_model, self._psf.device, mode="train"
                )
            else:
                process_function = get_drunet_function_v2(
                    process_model, self._psf.device, mode="train"
                )
        elif process is not None:
            # Otherwise, we assume it is a function.
            assert callable(process), "pre_process must be a callable function"
            process_function = process
            process_model = None
        else:
            process_function = None
            process_model = None

        if process_function is not None:
            process_param = torch.nn.Parameter(torch.tensor([1.0], device=self._psf.device))
        else:
            process_param = None

        return process_function, process_model, process_param

    def set_pre_process(self, pre_process):
        (
            self.pre_process,
            self.pre_process_model,
            self.pre_process_param,
        ) = self._prepare_process_block(pre_process)

    def set_post_process(self, post_process):
        (
            self.post_process,
            self.post_process_model,
            self.post_process_param,
        ) = self._prepare_process_block(post_process)

    def freeze_pre_process(self):
        """
        Method for freezing the pre process block.
        """
        if self.pre_process_param is not None:
            self.pre_process_param.requires_grad = False
        if self.pre_process_model is not None:
            for param in self.pre_process_model.parameters():
                param.requires_grad = False

    def freeze_post_process(self):
        """
        Method for freezing the post process block.
        """
        if self.post_process_param is not None:
            self.post_process_param.requires_grad = False
        if self.post_process_model is not None:
            for param in self.post_process_model.parameters():
                param.requires_grad = False

    def unfreeze_pre_process(self):
        """
        Method for unfreezing the pre process block.
        """
        if self.pre_process_param is not None:
            self.pre_process_param.requires_grad = True
        if self.pre_process_model is not None:
            for param in self.pre_process_model.parameters():
                param.requires_grad = True

    def unfreeze_post_process(self):
        """
        Method for unfreezing the post process block.
        """
        if self.post_process_param is not None:
            self.post_process_param.requires_grad = True
        if self.post_process_model is not None:
            for param in self.post_process_model.parameters():
                param.requires_grad = True

    def batch_call(self, batch):
        """
        Method for performing iterative reconstruction on a batch of images.
        This implementation is a properly vectorized implementation of FISTA.

        Parameters
        ----------
        batch : :py:class:`~torch.Tensor` of shape (batch, depth, channels, height, width)
            The lensless images to reconstruct.

        Returns
        -------
        :py:class:`~torch.Tensor` of shape (batch, depth, channels, height, width)
            The reconstructed images.
        """
        self._data = batch
        assert len(self._data.shape) == 5, "batch must be of shape (N, D, C, H, W)"
        batch_size = batch.shape[0]

        # pre process data
        if self.pre_process is not None:
            self._data = self.pre_process(self._data, self.pre_process_param)

        self.reset(batch_size=batch_size)

        # unrolled algorithm
        if not self.skip_unrolled:
            for i in range(self._n_iter):
                self._update(i)
            image_est = self._form_image()
        else:
            image_est = self._data

        # post process data
        if self.post_process is not None:
            final_est = self.post_process(image_est, self.post_process_param)
        else:
            final_est = image_est

        if self.return_unrolled_output:
            return final_est, image_est
        else:
            return final_est

    def apply(
        self,
        disp_iter=10,
        plot_pause=0.2,
        plot=True,
        save=False,
        gamma=None,
        ax=None,
        reset=True,
        output_intermediate=False,
    ):
        """
        Method for performing iterative reconstruction. Contrary to non-trainable reconstruction
        algorithm, the number of iteration isn't required. Note that `set_data` must be called
        beforehand.

        Parameters
        ----------
        disp_iter : int
            How often to display and/or intermediate reconstruction (in number
            of iterations). If `None` OR `plot` or `save` are False, no
            intermediate reconstruction will be plotted/saved.
        plot_pause : float
            Number of seconds to pause after displaying reconstruction.
        plot : bool
            Whether to plot final result, and intermediate results if
            `disp_iter` is not None.
        save : bool
            Whether to save final result (as PNG), and intermediate results if
            `disp_iter` is not None.
        gamma : float, optional
            Gamma correction factor to apply for plots. Default is None.
        ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill for plotting/saving, default is to create one.
        output_intermediate : bool, optional
            Whether to output intermediate reconstructions after preprocessing and before postprocessing.

        Returns
        -------
        final_im : :py:class:`~torch.Tensor`
            Final reconstruction.
        ax : :py:class:`~matplotlib.axes.Axes`
            `Axes` object on which final reconstruction is displayed. Only
            returning if `plot` or `save` is True.

        """
        pre_processed_image = None
        if self.pre_process is not None:
            self._data = self.pre_process(self._data, self.pre_process_param)
            if output_intermediate:
                pre_processed_image = self._data[0, ...].clone()

        if not self.skip_unrolled:
            im = super(TrainableReconstructionAlgorithm, self).apply(
                n_iter=self._n_iter,
                disp_iter=disp_iter,
                plot_pause=plot_pause,
                plot=plot,
                save=save,
                gamma=gamma,
                ax=ax,
                reset=reset,
            )
        else:
            im = self._data

        # remove plot if returned
        if plot:
            im, _ = im

        # post process data
        pre_post_process_image = None
        if self.post_process is not None:
            # apply post process
            if output_intermediate:
                pre_post_process_image = im.clone()
            im = self.post_process(im, self.post_process_param)

        if plot:
            ax = plot_image(self._get_numpy_data(im[0]), ax=ax, gamma=gamma)
            ax.set_title(
                "Final reconstruction after {} iterations and post process".format(self._n_iter)
            )
            if save:
                plt.savefig(plib.Path(save) / "final.png")

        if output_intermediate:
            return im, pre_processed_image, pre_post_process_image
        elif plot:
            return im, ax
        else:
            return im



# #############################################################################
# plot.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


import numpy as np
import warnings
import matplotlib.pyplot as plt

from lensless.utils.image import FLOAT_DTYPES, get_max_val, gamma_correction, autocorr2d


def plot_image(img, ax=None, gamma=None, normalize=True):
    """
    Plot image data.

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        Data to plot.
    ax : :py:class:`~matplotlib.axes.Axes`, optional
        `Axes` object to fill for plotting/saving, default is to create one.
    gamma : float, optional
        Gamma correction factor to apply for plots. Default is None.
    normalize : bool, optional
        Whether to normalize data to maximum range. Default is True.

    Returns
    -------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which image is plot.
    """

    # if we have only 1 depth, remove the axis
    if img.shape[0] == 1:
        img = img[0]

    # if we have only 1 color channel, remove the axis
    if img.shape[-1] == 1:
        img = img[..., 0]

    disp_img = None
    cmap = None

    # full 3D RGB format : [depth, width, height, color]
    is_3d = False
    if len(img.shape) == 4:
        disp_img = [np.sum(img, axis=axis) for axis in range(3)]
        cmap = None
        is_3d = True

    # data of length 3 means we have to infer whichever depth or color is missing, based on shape.
    elif len(img.shape) == 3:
        if img.shape[2] == 3:  # 2D rgb
            disp_img = [img]
            cmap = None

        else:  # 3D grayscale
            disp_img = [np.sum(img, axis=axis) for axis in range(3)]
            cmap = "gray"
            is_3d = True

    # data of length 2 means we have only width and height
    elif len(img.shape) == 2:  # 2D grayscale
        disp_img = [img]
        cmap = "gray"

    else:
        raise ValueError(f"Unexpected data shape : {img.shape}")

    max_val = [d.max() for d in disp_img]

    if not normalize:
        for i in range(len(max_val)):
            if disp_img[i].dtype not in FLOAT_DTYPES:
                max_val[i] = get_max_val(disp_img[i])
            else:
                max_val[i] = 1

    assert len(disp_img) == 1 or len(disp_img) == 3

    # need float image for gamma correction and plotting
    img_norm = disp_img.copy()
    for i in range(len(img_norm)):
        img_norm[i] = disp_img[i] / max_val[i]
        if gamma and gamma > 1:
            img_norm[i] = gamma_correction(img_norm[i], gamma=gamma)

    if ax is None:
        if not is_3d:
            _, ax = plt.subplots()
        else:
            _, ax = plt.subplots(2, 2)
    else:
        if is_3d:
            assert len(ax) == 2
            assert len(ax[0]) == 2

    if len(img_norm) == 1:
        ax.imshow(img_norm[0], cmap=cmap)

    else:

        # plot each projection separately
        ax[0, 0].imshow(img_norm[0], cmap=cmap)
        ax[0, 1].imshow(np.swapaxes(img_norm[2], 0, 1), cmap=cmap)
        ax[1, 0].imshow(img_norm[1], cmap=cmap)
        ax[1, 1].axis("off")
        ax[0, 1].set_xlabel("Depth")
        ax[1, 0].set_ylabel("Depth")

    return ax


def pixel_histogram(img, nbits=None, ax=None, log_scale=True):
    """
    Plot pixel value histogram.

    Parameters
    ----------
    img : py:class:`~numpy.ndarray`
        2D or 3D image.
    nbits : int, optional
        Bit-depth of camera data.
    ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill, default is to create one.
    log_scale : bool, optional
        Whether to use log scale in counting number of pixels.

    Return
    ------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which histogram is plot.
    """
    if ax is None:
        _, ax = plt.subplots()

    if nbits:
        # max_val = get_max_val(img, nbits)
        max_val = 2**nbits - 1
    else:
        max_val = int(img.max())

    if len(img.shape) == 3:
        # 3D image
        color_order = ("r", "g", "b")
        for i, col in enumerate(color_order):
            hist, bins = np.histogram(img[:, :, i].ravel(), bins=max_val, range=[0, max_val + 1])
            ax.plot(hist, color=col)
    else:
        # 2D image
        vals = img.flatten()
        hist, bins = np.histogram(vals, bins=max_val, range=[0, max_val + 1])
        ax.plot(hist, color="gray")
    ax.set_xlim([max_val - 1.1 * max_val, max_val * 1.1])

    if log_scale:
        ax.set_yscale("log")
    ax.set_xlabel("Pixel value")
    ax.grid()

    return ax


def plot_cross_section(
    vals,
    idx=None,
    ax=None,
    dB=True,
    plot_db_drop=3,
    min_val=1e-4,
    max_val=None,
    plot_width=None,
    **kwargs,
):
    """
    Plot cross-section of a 2-D image.

    Parameters
    ----------
    vals : py:class:`~numpy.ndarray`
        2-D image data.
    idx : int, optional
        Row for which to plot cross-section. Default is to take middle.
    ax : :py:class:`~matplotlib.axes.Axes`, optional
        `Axes` object to fill, default is to create one.
    dB : bool, optional
        Whether to plot in dB scale.

    Return
    ------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which cross-section is plot.
    """

    if ax is None:
        _, ax = plt.subplots()

    # get cross-section
    if idx is None:
        # if no index, take cross-section with maximum value
        max_idx = np.unravel_index(np.argmax(vals, axis=None), vals.shape)
        idx = max_idx[0]

    cross_section = vals[idx, :].astype(np.float32)

    # normalize
    if max_val is None:
        max_val = cross_section.max()
    cross_section /= max_val
    min_val = max(min_val, cross_section.min())

    if dB:
        cross_section[cross_section < min_val] = min_val
        cross_section = 10 * np.log10(cross_section)
        min_val = 10 * np.log10(min_val)
        ax.set_ylabel("dB")
    x_vals = np.arange(len(cross_section))
    x_vals -= np.argmax(cross_section)
    ax.plot(x_vals, cross_section, **kwargs)
    ax.set_ylim([min_val, 0])
    if plot_width is not None:
        half_width = plot_width // 2 + 1
        ax.set_xlim([-half_width, half_width])
    ax.grid()

    if dB and plot_db_drop:
        cross_section -= np.max(cross_section)
        zero_crossings = np.where(np.diff(np.signbit(cross_section + plot_db_drop)))[0]
        if len(zero_crossings) >= 2:
            zero_crossings -= np.argmax(cross_section)
            width = zero_crossings[-1] - zero_crossings[0]
            ax.set_title(f"-{plot_db_drop}dB width = {width}")
            ax.axvline(x=zero_crossings[0], c="k", linestyle="--")
            ax.axvline(x=zero_crossings[-1], c="k", linestyle="--")
        else:
            warnings.warn(
                "Width could not be determined. Did not detect two -{} points : {}".format(
                    plot_db_drop, zero_crossings
                )
            )

    return ax, cross_section


def plot_autocorr2d(vals, pad_mode="reflect", ax=None):
    """
    Plot 2-D autocorrelation of image.

    Parameters
    ----------
    vals : py:class:`~numpy.ndarray`
        2-D image.
    pad_mode : str
        Desired padding. See NumPy documentation: https://numpy.org/doc/stable/reference/generated/numpy.pad.html
    ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill, default is to create one.

    Return
    ------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which auto-correlation is plot.
    autocorr : py:class:`~numpy.ndarray`
        Auto-correlation.
    """

    nbit_plot = 8
    max_val_plot = 2**nbit_plot - 1

    # compute autocorrelation
    autocorr = autocorr2d(vals, pad_mode=pad_mode)

    # rescale for plotting
    data = autocorr - np.min(autocorr)
    data = data / np.max(np.abs(data))  # normalize the data to 0 - 1
    data = max_val_plot * data  # Now scale by bit depth
    autocorr_img = data.astype(np.uint8)

    if ax is None:
        _, ax = plt.subplots()

    ax.imshow(autocorr_img, cmap="gray", vmin=0, vmax=max_val_plot)
    ax.axis("off")
    return ax, autocorr



# #############################################################################
# mask.py
# =================
# Authors :
# Aaron FARGEON [aa.fargeon@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Mask
====

This module provides utilities to create different types of masks (:py:class:`~lensless.hardware.mask.CodedAperture`,
:py:class:`~lensless.hardware.mask.PhaseContour`,
:py:class:`~lensless.hardware.mask.FresnelZoneAperture`) and simulate the corresponding PSF.

"""


import abc
import warnings
import numpy as np
import cv2 as cv
from math import sqrt
from perlin_numpy import generate_perlin_noise_2d
from sympy.ntheory import quadratic_residues
from scipy.signal import max_len_seq
from scipy.linalg import circulant
from numpy.linalg import multi_dot
from waveprop.fresnel import fresnel_conv
from waveprop.rs import angular_spectrum
from waveprop.noise import add_shot_noise
from lensless.hardware.sensor import VirtualSensor
from lensless.utils.image import resize

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class Mask(abc.ABC):
    """
    Parent ``Mask`` class. Attributes common to each type of mask.
    """

    def __init__(
        self,
        resolution,
        distance_sensor,
        size=None,
        feature_size=None,
        psf_wavelength=[460e-9, 550e-9, 640e-9],
        **kwargs
    ):
        """
        Constructor from parameters of the user's choice.

        Parameters
        ----------
        resolution: array_like
            Resolution of the  mask (px).
        distance_sensor: float
            Distance between the mask and the sensor (m).
        size: array_like
            Size of the sensor (m). Only one of ``size`` or ``feature_size`` needs to be specified.
        feature_size: float or array_like
            Size of the feature (m). Only one of ``size`` or ``feature_size`` needs to be specified.
        psf_wavelength: list, optional
            List of wavelengths to simulate PSF (m). Default is [460e-9, 550e-9, 640e-9] nm (blue, green, red).
        """

        resolution = np.array(resolution)
        assert len(resolution) == 2, "Sensor resolution should be of length 2"

        assert (
            size is not None or feature_size is not None
        ), "Either sensor_size or feature_size should be specified"
        if size is None:
            size = np.array(resolution * feature_size)
        else:
            size = np.array(size)
            assert len(size) == 2, "Sensor size should be of length 2"
        if feature_size is None:
            feature_size = np.array(size) / np.array(resolution)
        else:
            if isinstance(feature_size, float):
                feature_size = np.array([feature_size, feature_size])
            else:
                assert len(feature_size) == 2, "Feature size should be of length 2"
                feature_size = np.array(feature_size)
            assert np.all(feature_size > 0), "Feature size should be positive"
        assert np.all(resolution * feature_size <= size)

        self.phase_mask = None
        self.resolution = resolution
        self.size = size
        if feature_size is None:
            self.feature_size = self.size / self.resolution
        else:
            self.feature_size = feature_size
        self.distance_sensor = distance_sensor

        # create mask
        self.mask = None
        self.create_mask()
        self.shape = self.mask.shape

        # PSF
        self.psf_wavelength = psf_wavelength
        self.psf = None
        self.compute_psf()

    @classmethod
    def from_sensor(cls, sensor_name, downsample=None, **kwargs):
        """
        Constructor from an existing virtual sensor that copies over the sensor parameters
        (sensor resolution, sensor size, feature size).

        Parameters
        ----------
        sensor_name: str
            Name of the sensor. See :py:class:`~lensless.hardware.sensor.SensorOptions`.
        downsample: float, optional
            Downsampling factor.
        **kwargs:
            Additional arguments for the mask constructor. See the abstract class :py:class:`~lensless.hardware.mask.Mask`
            and the corresponding subclass for more details.

        Example
        -------

        .. code-block:: python

            mask = CodedAperture.from_sensor(sensor_name=SensorOptions.RPI_HQ, downsample=8, ...)
        """
        sensor = VirtualSensor.from_name(sensor_name, downsample)
        return cls(
            resolution=tuple(sensor.resolution.copy()),
            size=tuple(sensor.size.copy()),
            feature_size=sensor.pixel_size.copy(),
            **kwargs
        )

    @abc.abstractmethod
    def create_mask(self):
        """
        Abstract mask creation method that creates mask with subclass-specific function.
        """
        pass

    def compute_psf(self):
        """
        Compute the intensity PSF with bandlimited angular spectrum (BLAS) for each wavelength.
        Common to all types of masks.
        """
        psf = np.zeros(tuple(self.resolution) + (len(self.psf_wavelength),), dtype=np.complex64)
        for i, wv in enumerate(self.psf_wavelength):
            psf[:, :, i] = angular_spectrum(
                u_in=self.mask,
                wv=wv,
                d1=self.feature_size,
                dz=self.distance_sensor,
                dtype=np.float32,
                bandlimit=True,
            )[0]

        # intensity PSF
        self.psf = np.abs(psf) ** 2


class CodedAperture(Mask):
    """
    Coded aperture mask as in `FlatCam <https://arxiv.org/abs/1509.00116>`_.
    """

    def __init__(self, method="MLS", n_bits=8, **kwargs):
        """
        Coded aperture mask contructor (FlatCam).

        Parameters
        ----------
        method: str
            Pattern generation method (MURA or MLS). Default is ``MLS``.
        n_bits: int, optional
            Number of bits for pattern generation.
            Size is ``4*n_bits + 1`` for MURA and ``2^n - 1`` for MLS.
            Default is 8 (for a 255x255 MLS mask).
        **kwargs:
            The keyword arguments are passed to the parent class :py:class:`~lensless.hardware.mask.Mask`.
        """

        self.row = None
        self.col = None
        self.method = method
        self.n_bits = n_bits

        super().__init__(**kwargs)

    def create_mask(self):
        """
        Creating coded aperture mask using either the MURA of MLS method.
        """
        assert self.method.upper() in ["MURA", "MLS"], "Method should be either 'MLS' or 'MURA'"

        # Generating pattern
        if self.method.upper() == "MURA":
            self.mask = self.squarepattern(4 * self.n_bits + 1)[1:, 1:]
            self.row = 2 * self.mask[0, :] - 1
            self.col = 2 * self.mask[:, 0] - 1
        else:
            seq = max_len_seq(self.n_bits)[0] * 2 - 1
            h_r = np.r_[seq, seq]
            self.row = h_r
            self.col = h_r
            self.mask = (np.outer(h_r, h_r) + 1) / 2

        # Upscaling
        if np.any(self.resolution != self.mask.shape):
            upscaled_mask = resize(
                self.mask[:, :, np.newaxis], shape=tuple(self.resolution) + (1,)
            ).squeeze()
            upscaled_mask = np.clip(upscaled_mask, 0, 1)
            self.mask = np.round(upscaled_mask).astype(int)

    def is_prime(self, n):
        """
        Assess whether a number is prime or not.

        Parameters
        ----------
        n: int
            The number we want to check.
        """
        if n % 2 == 0 and n > 2:
            return False
        return all(n % i for i in range(3, int(sqrt(n)) + 1, 2))

    def squarepattern(self, p):
        """
        Generate MURA square pattern.

        Parameters
        ----------
        p: int
            Number of bits.
        """
        if not self.is_prime(p):
            raise ValueError("p is not a valid length. It must be prime.")
        A = np.zeros((p, p), dtype=int)
        q = quadratic_residues(p)
        A[1:, 0] = 1
        for j in range(1, p):
            for i in range(1, p):
                if not ((i - 1 in q) != (j - 1 in q)):
                    A[i, j] = 1
        return A

    def get_conv_matrices(self, img_shape):
        """
        Get theoretical left and right convolution matrices for the separable mask.

        Such that measurement model is given ``P @ img @ Q.T``.

        Parameters
        ----------
        img_shape: tuple
            Shape of the image to being convolved.

        Returns
        -------
        P: :py:class:`~numpy.ndarray`
            Left convolution matrix.
        Q: :py:class:`~numpy.ndarray`
            Right convolution matrix.

        """

        P = circulant(np.resize(self.col, self.resolution[0]))[:, : img_shape[0]]
        Q = circulant(np.resize(self.row, self.resolution[1]))[:, : img_shape[1]]

        return P, Q

    def simulate(self, obj, snr_db=20):
        """
        Simulate the mask measurement of an image. Apply left and right convolution matrices,
        add noise and return the measurement.

        Parameters
        ----------
        obj: :py:class:`~numpy.ndarray`
            Image to simulate.
        snr_db: float, optional
            Signal-to-noise ratio (dB) of the simulated measurement. Default is 20 dB.
        """
        assert len(obj.shape) == 3, "Object should be a 3D array (HxWxC) even if grayscale."

        # Get convolution matrices
        P, Q = self.get_conv_matrices(obj.shape)

        # Convolve image
        n_channels = obj.shape[-1]

        if torch_available and isinstance(obj, torch.Tensor):
            P = torch.from_numpy(P).float()
            Q = torch.from_numpy(Q).float()
            meas = torch.dstack(
                [torch.linalg.multi_dot([P, obj[:, :, c], Q.T]) for c in range(n_channels)]
            ).float()
        else:
            meas = np.dstack([multi_dot([P, obj[:, :, c], Q.T]) for c in range(n_channels)])

        # Add noise
        if snr_db is not None:
            meas = add_shot_noise(meas, snr_db=snr_db)

        if torch_available and isinstance(obj, torch.Tensor):
            meas = meas.to(obj)

        return meas


class PhaseContour(Mask):
    """
    Phase contour mask as in `PhlatCam <https://ieeexplore.ieee.org/document/9076617>`_.
    """

    def __init__(
        self, noise_period=(16, 16), refractive_index=1.2, n_iter=10, design_wv=532e-9, **kwargs
    ):
        """
        Phase contour mask contructor (PhlatCam).

        Parameters
        ----------
        noise_period: array_like, optional
            Noise period of the Perlin noise (px). Default is (8, 8).
        refractive_index: float, optional
            Refractive index of the mask substrate. Default is 1.2.
        n_iter: int, optional
            Number of iterations for the phase retrieval algorithm. Default is 10.
        design_wv: float, optional
            Wavelength used to design the mask (m). Default is 532e-9, as in the PhlatCam paper.
        **kwargs:
            The keyword arguments are passed to the parent class :py:class:`~lensless.hardware.mask.Mask`.
        """

        self.target_psf = None
        self.phase_pattern = None
        self.height_map = None
        self.noise_period = noise_period
        self.refractive_index = refractive_index
        self.n_iter = n_iter
        self.design_wv = design_wv

        super().__init__(**kwargs)

    def create_mask(self):
        """
        Creating phase contour from edges of Perlin noise.
        """

        # Creating Perlin noise
        proper_dim_1 = (self.resolution[0] // self.noise_period[0]) * self.noise_period[0]
        proper_dim_2 = (self.resolution[1] // self.noise_period[1]) * self.noise_period[1]
        noise = generate_perlin_noise_2d((proper_dim_1, proper_dim_2), self.noise_period)

        # Upscaling to correspond to sensor size
        if np.any(self.resolution != noise.shape):
            noise = resize(noise[:, :, np.newaxis], shape=tuple(self.resolution) + (1,)).squeeze()

        # Edge detection
        binary = np.clip(np.round(np.interp(noise, (-1, 1), (0, 1))), a_min=0, a_max=1)
        self.target_psf = cv.Canny(np.interp(binary, (-1, 1), (0, 255)).astype(np.uint8), 0, 255)

        # Computing mask and height map
        phase_mask, height_map = phase_retrieval(
            target_psf=self.target_psf,
            wv=self.design_wv,
            d1=self.feature_size,
            dz=self.distance_sensor,
            n=self.refractive_index,
            n_iter=self.n_iter,
            height_map=True,
        )
        self.height_map = height_map
        self.phase_pattern = phase_mask
        self.mask = np.exp(1j * phase_mask)


def phase_retrieval(target_psf, wv, d1, dz, n=1.2, n_iter=10, height_map=False):
    """
    Iterative phase retrieval algorithm similar to `PhlatCam <https://ieeexplore.ieee.org/document/9076617>`_,
    using Fresnel propagation.

    Parameters
    ----------
    target_psf: array_like
        Target PSF to optimize the phase mask for.
    wv: float
        Wavelength (m).
    d1: float
        Sample period on the sensor i.e. pixel size (m).
    dz: float
        Propagation distance between the mask and the sensor.
    n: float
        Refractive index of the mask substrate. Default is 1.2.
    n_iter: int
        Number of iterations. Default value is 10.
    """
    M_p = np.sqrt(target_psf)

    if hasattr(d1, "__len__"):
        if d1[0] != d1[1]:
            warnings.warn("Non-square pixel, first dimension taken as feature size.")
        d1 = d1[0]

    for _ in range(n_iter):
        # back propagate from sensor to mask
        M_phi = fresnel_conv(M_p, wv, d1, -dz, dtype=np.float32)[0]
        # constrain amplitude at mask to be unity, i.e. phase pattern
        M_phi = np.exp(1j * np.angle(M_phi))
        # forward propagate from mask to sensor
        M_p = fresnel_conv(M_phi, wv, d1, dz, dtype=np.float32)[0]
        # constrain amplitude to be sqrt(PSF)
        M_p = np.sqrt(target_psf) * np.exp(1j * np.angle(M_p))

    phi = (np.angle(M_phi) + 2 * np.pi) % (2 * np.pi)

    if height_map:
        return phi, wv * phi / (2 * np.pi * (n - 1))
    else:
        return phi


class FresnelZoneAperture(Mask):
    """
    Fresnel Zone Aperture (FZA) mask as in `this work <https://www.nature.com/articles/s41377-020-0289-9>`_,
    namely binarized cosine function.
    """

    def __init__(self, radius=0.32e-3, **kwargs):
        """
        Fresnel Zone Aperture mask contructor.

        Parameters
        ----------
        radius: float
            characteristic radius of the FZA (m)
            default value: 5e-4
        **kwargs:
            The keyword arguments are passed to the parent class :py:class:`~lensless.hardware.mask.Mask`.
        """

        self.radius = radius

        super().__init__(**kwargs)

    def create_mask(self):
        """
        Creating binary Fresnel Zone Aperture mask.
        """
        dim = self.resolution
        x, y = np.meshgrid(
            np.linspace(-dim[1] / 2, dim[1] / 2 - 1, dim[1]),
            np.linspace(-dim[0] / 2, dim[0] / 2 - 1, dim[0]),
        )
        radius_px = self.radius / self.feature_size[0]
        mask = 0.5 * (1 + np.cos(np.pi * (x**2 + y**2) / radius_px**2))
        self.mask = np.round(mask)



# #############################################################################
# train_unrolled.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Train unrolled version of reconstruction algorithm.

```
python scripts/recon/train_unrolled.py
```

By default it uses the configuration from the file `configs/train_unrolledADMM.yaml`.

To train pre- and post-processing networks, use the following command:
```
python scripts/recon/train_unrolled.py -cn train_pre-post-processing
```

To fine-tune the DiffuserCam PSF, use the following command:
```
python scripts/recon/train_unrolled.py -cn fine-tune_PSF
```

To train a PSF from scratch with a simulated dataset, use the following command:
```
python scripts/recon/train_unrolled.py -cn train_psf_from_scratch
```

"""

import logging
import hydra
from hydra.utils import get_original_cwd
import os
import numpy as np
import time
from lensless import UnrolledFISTA, UnrolledADMM, TrainableInversion
from lensless.utils.dataset import (
    DiffuserCamMirflickr,
    SimulatedFarFieldDataset,
    SimulatedDatasetTrainableMask,
    DigiCamCelebA,
    HITLDatasetTrainableMask,
)
from torch.utils.data import Subset
import lensless.hardware.trainable_mask
from lensless.hardware.slm import full2subpattern
from lensless.hardware.sensor import VirtualSensor
from lensless.recon.utils import create_process_network
from lensless.utils.image import rgb2gray, is_grayscale
from lensless.utils.simulation import FarFieldSimulator
from lensless.recon.utils import Trainer
import torch
from torchvision import transforms, datasets
from lensless.utils.io import load_psf
from lensless.utils.io import save_image
from lensless.utils.plot import plot_image
from lensless import ADMM
import matplotlib.pyplot as plt

# A logger for this file
log = logging.getLogger(__name__)


def simulate_dataset(config, generator=None):

    if config.torch_device == "cuda" and torch.cuda.is_available():
        device = "cuda"
    else:
        device = "cpu"

    # -- prepare PSF
    psf = None
    if config.trainable_mask.mask_type is None or config.trainable_mask.initial_value == "psf":
        psf_fp = os.path.join(get_original_cwd(), config.files.psf)
        psf, _ = load_psf(
            psf_fp,
            downsample=config.files.downsample,
            return_float=True,
            return_bg=True,
            bg_pix=(0, 15),
        )
        if config.files.diffusercam_psf:
            transform_BRG2RGB = transforms.Lambda(lambda x: x[..., [2, 1, 0]])
            psf = transform_BRG2RGB(torch.from_numpy(psf))

        # drop depth dimension
        psf = psf.to(device)

    else:
        # training mask / PSF
        mask = prep_trainable_mask(config, psf)
        psf = mask.get_psf().to(device)

    # -- load dataset
    pre_transform = None
    transforms_list = [transforms.ToTensor()]
    data_path = os.path.join(get_original_cwd(), "data")
    if config.simulation.grayscale:
        transforms_list.append(transforms.Grayscale())

    if config.files.dataset == "mnist":
        transform = transforms.Compose(transforms_list)
        train_ds = datasets.MNIST(root=data_path, train=True, download=True, transform=transform)
        test_ds = datasets.MNIST(root=data_path, train=False, download=True, transform=transform)
    elif config.files.dataset == "fashion_mnist":
        transform = transforms.Compose(transforms_list)
        train_ds = datasets.FashionMNIST(
            root=data_path, train=True, download=True, transform=transform
        )
        test_ds = datasets.FashionMNIST(
            root=data_path, train=False, download=True, transform=transform
        )
    elif config.files.dataset == "cifar10":
        transform = transforms.Compose(transforms_list)
        train_ds = datasets.CIFAR10(root=data_path, train=True, download=True, transform=transform)
        test_ds = datasets.CIFAR10(root=data_path, train=False, download=True, transform=transform)
    elif config.files.dataset == "CelebA":
        root = config.files.celeba_root
        data_path = os.path.join(root, "celeba")
        assert os.path.isdir(
            data_path
        ), f"Data path {data_path} does not exist. Make sure you download the CelebA dataset and provide the parent directory as 'config.files.celeba_root'. Download link: https://mmlab.ie.cuhk.edu.hk/projects/CelebA.html"
        transform = transforms.Compose(transforms_list)
        if config.files.n_files is None:
            train_ds = datasets.CelebA(
                root=root, split="train", download=False, transform=transform
            )
            test_ds = datasets.CelebA(root=root, split="test", download=False, transform=transform)
        else:
            ds = datasets.CelebA(root=root, split="all", download=False, transform=transform)

            ds = Subset(ds, np.arange(config.files.n_files))

            train_size = int((1 - config.files.test_size) * len(ds))
            test_size = len(ds) - train_size
            train_ds, test_ds = torch.utils.data.random_split(
                ds, [train_size, test_size], generator=generator
            )
    else:
        raise NotImplementedError(f"Dataset {config.files.dataset} not implemented.")

    # convert PSF
    if config.simulation.grayscale and not is_grayscale(psf):
        psf = rgb2gray(psf)

    # check if gpu is available
    device_conv = config.torch_device
    if device_conv == "cuda" and torch.cuda.is_available():
        device_conv = "cuda"
    else:
        device_conv = "cpu"

    # create simulator
    simulator = FarFieldSimulator(
        psf=psf,
        is_torch=True,
        **config.simulation,
    )

    # create Pytorch dataset and dataloader
    crop = config.files.crop.copy() if config.files.crop is not None else None
    if mask is None:
        train_ds_prop = SimulatedFarFieldDataset(
            dataset=train_ds,
            simulator=simulator,
            dataset_is_CHW=True,
            device_conv=device_conv,
            flip=config.simulation.flip,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            crop=crop,
            downsample=config.files.downsample,
            pre_transform=pre_transform,
        )
        test_ds_prop = SimulatedFarFieldDataset(
            dataset=test_ds,
            simulator=simulator,
            dataset_is_CHW=True,
            device_conv=device_conv,
            flip=config.simulation.flip,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            crop=crop,
            downsample=config.files.downsample,
            pre_transform=pre_transform,
        )
    else:
        if config.measure is not None:

            train_ds_prop = HITLDatasetTrainableMask(
                rpi_username=config.measure.rpi_username,
                rpi_hostname=config.measure.rpi_hostname,
                celeba_root=config.files.celeba_root,
                display_config=config.measure.display,
                capture_config=config.measure.capture,
                mask_center=config.trainable_mask.ap_center,
                dataset=train_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )

            test_ds_prop = HITLDatasetTrainableMask(
                rpi_username=config.measure.rpi_username,
                rpi_hostname=config.measure.rpi_hostname,
                celeba_root=config.files.celeba_root,
                display_config=config.measure.display,
                capture_config=config.measure.capture,
                mask_center=config.trainable_mask.ap_center,
                dataset=test_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )

        else:

            train_ds_prop = SimulatedDatasetTrainableMask(
                dataset=train_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )
            test_ds_prop = SimulatedDatasetTrainableMask(
                dataset=test_ds,
                mask=mask,
                simulator=simulator,
                dataset_is_CHW=True,
                device_conv=device_conv,
                flip=config.simulation.flip,
                vertical_shift=config.files.vertical_shift,
                horizontal_shift=config.files.horizontal_shift,
                crop=crop,
                downsample=config.files.downsample,
                pre_transform=pre_transform,
            )

    return train_ds_prop, test_ds_prop, mask


def prep_trainable_mask(config, psf=None, downsample=None):
    mask = None
    color_filter = None
    downsample = config.files.downsample if downsample is None else downsample
    if config.trainable_mask.mask_type is not None:
        mask_class = getattr(lensless.hardware.trainable_mask, config.trainable_mask.mask_type)

        if config.trainable_mask.initial_value == "random":
            if psf is not None:
                initial_mask = torch.rand_like(psf)
            else:
                sensor = VirtualSensor.from_name(config.simulation.sensor, downsample=downsample)
                resolution = sensor.resolution
                initial_mask = torch.rand((1, *resolution, 3))
        elif config.trainable_mask.initial_value == "psf":
            initial_mask = psf.clone()
        # if file ending with "npy"
        elif config.trainable_mask.initial_value.endswith("npy"):
            pattern = np.load(os.path.join(get_original_cwd(), config.trainable_mask.initial_value))

            initial_mask = full2subpattern(
                pattern=pattern,
                shape=config.trainable_mask.ap_shape,
                center=config.trainable_mask.ap_center,
                slm=config.trainable_mask.slm,
            )
            initial_mask = torch.from_numpy(initial_mask.astype(np.float32))

            # prepare color filter if needed
            from waveprop.devices import slm_dict
            from waveprop.devices import SLMParam as SLMParam_wp

            slm_param = slm_dict[config.trainable_mask.slm]
            if (
                config.trainable_mask.train_color_filter
                and SLMParam_wp.COLOR_FILTER in slm_param.keys()
            ):
                color_filter = slm_param[SLMParam_wp.COLOR_FILTER]
                color_filter = torch.from_numpy(color_filter.copy()).to(dtype=torch.float32)

                # add small random values
                color_filter = color_filter + 0.1 * torch.rand_like(color_filter)
        else:
            raise ValueError(
                f"Initial PSF value {config.trainable_mask.initial_value} not supported"
            )

        if config.trainable_mask.grayscale and not is_grayscale(initial_mask):
            initial_mask = rgb2gray(initial_mask)

        mask = mask_class(
            initial_mask,
            optimizer="Adam",
            downsample=downsample,
            color_filter=color_filter,
            **config.trainable_mask,
        )

    return mask


@hydra.main(version_base=None, config_path="../../configs", config_name="train_unrolledADMM")
def train_unrolled(config):

    # set seed
    seed = config.seed
    torch.manual_seed(seed)
    np.random.seed(seed)
    generator = torch.Generator().manual_seed(seed)

    if config.start_delay is not None:
        # wait for this time before starting script
        delay = config.start_delay * 60
        start_time = time.time() + delay
        start_time = time.strftime("%H:%M:%S", time.localtime(start_time))
        print(f"\nScript will start at {start_time}")
        time.sleep(delay)

    save = config.save
    if save:
        save = os.getcwd()

    if config.torch_device == "cuda" and torch.cuda.is_available():
        log.info("Using GPU for training.")
        device = "cuda"
    else:
        log.info("Using CPU for training.")
        device = "cpu"

    # load dataset and create dataloader
    train_set = None
    test_set = None
    psf = None
    crop = None
    if "DiffuserCam" in config.files.dataset:

        original_path = os.path.join(get_original_cwd(), config.files.dataset)
        psf_path = os.path.join(get_original_cwd(), config.files.psf)
        dataset = DiffuserCamMirflickr(
            dataset_dir=original_path,
            psf_path=psf_path,
            downsample=config.files.downsample,
            input_snr=config.files.input_snr,
        )
        dataset.psf = dataset.psf.to(device)
        # train-test split as in https://waller-lab.github.io/LenslessLearning/dataset.html
        # first 1000 files for test, the rest for training
        train_indices = dataset.allowed_idx[dataset.allowed_idx > 1000]
        test_indices = dataset.allowed_idx[dataset.allowed_idx <= 1000]
        if config.files.n_files is not None:
            train_indices = train_indices[: config.files.n_files]
            test_indices = test_indices[: config.files.n_files]

        train_set = Subset(dataset, train_indices)
        test_set = Subset(dataset, test_indices)

        # -- if learning mask
        mask = prep_trainable_mask(config, dataset.psf)
        if mask is not None:
            # plot initial PSF
            psf_np = mask.get_psf().detach().cpu().numpy()[0, ...]
            if config.trainable_mask.grayscale:
                psf_np = psf_np[:, :, -1]

            save_image(psf_np, os.path.join(save, "psf_initial.png"))
            plot_image(psf_np, gamma=config.display.gamma)
            plt.savefig(os.path.join(save, "psf_initial_plot.png"))

        psf = dataset.psf

    elif "celeba_adafruit" in config.files.dataset:

        dataset = DigiCamCelebA(
            data_dir=os.path.join(get_original_cwd(), config.files.dataset),
            celeba_root=config.files.celeba_root,
            psf_path=os.path.join(get_original_cwd(), config.files.psf),
            downsample=config.files.downsample,
            vertical_shift=config.files.vertical_shift,
            horizontal_shift=config.files.horizontal_shift,
            simulation_config=config.simulation,
            crop=config.files.crop,
            input_snr=config.files.input_snr,
        )
        crop = dataset.crop
        dataset.psf = dataset.psf.to(device)
        log.info(f"Data shape :  {dataset[0][0].shape}")

        if config.files.n_files is not None:
            dataset = Subset(dataset, np.arange(config.files.n_files))
            dataset.psf = dataset.dataset.psf

        # train-test split
        train_size = int((1 - config.files.test_size) * len(dataset))
        test_size = len(dataset) - train_size
        train_set, test_set = torch.utils.data.random_split(
            dataset, [train_size, test_size], generator=generator
        )

        # -- if learning mask
        downsample = config.files.downsample * 4  # measured files are 4x downsampled
        mask = prep_trainable_mask(config, dataset.psf, downsample=downsample)

        if mask is not None:
            # plot initial PSF
            with torch.no_grad():
                psf_np = mask.get_psf().detach().cpu().numpy()[0, ...]
                if config.trainable_mask.grayscale:
                    psf_np = psf_np[:, :, -1]

            save_image(psf_np, os.path.join(save, "psf_initial.png"))
            plot_image(psf_np, gamma=config.display.gamma)
            plt.savefig(os.path.join(save, "psf_initial_plot.png"))

            # save original PSF as well
            psf_meas = dataset.psf.detach().cpu().numpy()[0, ...]
            plot_image(psf_meas, gamma=config.display.gamma)
            plt.savefig(os.path.join(save, "psf_meas_plot.png"))

            with torch.no_grad():
                psf = mask.get_psf().to(dataset.psf)

        else:

            psf = dataset.psf

        # print info about PSF
        log.info(f"PSF shape : {psf.shape}")
        log.info(f"PSF min : {psf.min()}")
        log.info(f"PSF max : {psf.max()}")
        log.info(f"PSF dtype : {psf.dtype}")
        log.info(f"PSF norm : {psf.norm()}")

    else:

        train_set, test_set, mask = simulate_dataset(config, generator=generator)
        psf = train_set.psf
        crop = train_set.crop

    assert train_set is not None
    assert psf is not None

    # reconstruct lensless with ADMM
    with torch.no_grad():
        if config.test_idx is not None:

            log.info("Reconstruction a few images with ADMM...")

            for i, _idx in enumerate(config.test_idx):

                lensless, lensed = test_set[_idx]
                recon = ADMM(psf)

                recon.set_data(lensless.to(psf.device))
                res = recon.apply(disp_iter=None, plot=False, n_iter=10)
                res_np = res[0].cpu().numpy()
                res_np = res_np / res_np.max()

                lensed_np = lensed[0].cpu().numpy()

                lensless_np = lensless[0].cpu().numpy()
                save_image(lensless_np, f"lensless_raw_{_idx}.png")

                # -- plot lensed and res on top of each other
                if config.training.crop_preloss:
                    assert crop is not None

                    res_np = res_np[
                        crop["vertical"][0] : crop["vertical"][1],
                        crop["horizontal"][0] : crop["horizontal"][1],
                    ]
                    lensed_np = lensed_np[
                        crop["vertical"][0] : crop["vertical"][1],
                        crop["horizontal"][0] : crop["horizontal"][1],
                    ]
                    if i == 0:
                        log.info(f"Cropped shape :  {res_np.shape}")

                save_image(res_np, f"lensless_recon_{_idx}.png")
                save_image(lensed_np, f"lensed_{_idx}.png")

                plt.figure()
                plt.imshow(lensed_np, alpha=0.4)
                plt.imshow(res_np, alpha=0.7)
                plt.savefig(f"overlay_lensed_recon_{_idx}.png")

    log.info(f"Train test size : {len(train_set)}")
    log.info(f"Test test size : {len(test_set)}")

    start_time = time.time()

    # Load pre-process model
    pre_process, pre_process_name = create_process_network(
        config.reconstruction.pre_process.network,
        config.reconstruction.pre_process.depth,
        nc=config.reconstruction.pre_process.nc,
        device=device,
    )
    pre_proc_delay = config.reconstruction.pre_process.delay

    # Load post-process model
    post_process, post_process_name = create_process_network(
        config.reconstruction.post_process.network,
        config.reconstruction.post_process.depth,
        nc=config.reconstruction.post_process.nc,
        device=device,
    )
    post_proc_delay = config.reconstruction.post_process.delay

    if config.reconstruction.post_process.train_last_layer:
        for name, param in post_process.named_parameters():
            if "m_tail" in name:
                param.requires_grad = True
            else:
                param.requires_grad = False
            # print(name, param.requires_grad, param.numel())

    # initialize pre- and post processor with another model
    if config.reconstruction.init_processors is not None:
        from lensless.recon.model_dict import load_model, model_dict

        model_orig = load_model(
            model_dict["diffusercam"]["mirflickr"][config.reconstruction.init_processors],
            psf=psf,
            device=device,
        )

        # -- replace pre-process
        if config.reconstruction.init_pre:
            params1 = model_orig.pre_process_model.named_parameters()
            params2 = pre_process.named_parameters()
            dict_params2 = dict(params2)
            for name1, param1 in params1:
                if name1 in dict_params2:
                    dict_params2[name1].data.copy_(param1.data)

        # -- replace post-process
        if config.reconstruction.init_post:
            params1_post = model_orig.post_process_model.named_parameters()
            params2_post = post_process.named_parameters()
            dict_params2_post = dict(params2_post)
            for name1, param1 in params1_post:
                if name1 in dict_params2_post:
                    dict_params2_post[name1].data.copy_(param1.data)

    # create reconstruction algorithm
    if config.reconstruction.method == "unrolled_fista":
        recon = UnrolledFISTA(
            psf,
            n_iter=config.reconstruction.unrolled_fista.n_iter,
            tk=config.reconstruction.unrolled_fista.tk,
            pad=True,
            learn_tk=config.reconstruction.unrolled_fista.learn_tk,
            pre_process=pre_process if pre_proc_delay is None else None,
            post_process=post_process if post_proc_delay is None else None,
            skip_unrolled=config.reconstruction.skip_unrolled,
            return_unrolled_output=True if config.unrolled_output_factor > 0 else False,
        ).to(device)
    elif config.reconstruction.method == "unrolled_admm":
        recon = UnrolledADMM(
            psf,
            n_iter=config.reconstruction.unrolled_admm.n_iter,
            mu1=config.reconstruction.unrolled_admm.mu1,
            mu2=config.reconstruction.unrolled_admm.mu2,
            mu3=config.reconstruction.unrolled_admm.mu3,
            tau=config.reconstruction.unrolled_admm.tau,
            pre_process=pre_process if pre_proc_delay is None else None,
            post_process=post_process if post_proc_delay is None else None,
            skip_unrolled=config.reconstruction.skip_unrolled,
            return_unrolled_output=True if config.unrolled_output_factor > 0 else False,
        ).to(device)
    elif config.reconstruction.method == "trainable_inv":
        recon = TrainableInversion(
            psf,
            K=config.reconstruction.trainable_inv.K,
            pre_process=pre_process if pre_proc_delay is None else None,
            post_process=post_process if post_proc_delay is None else None,
            return_unrolled_output=True if config.unrolled_output_factor > 0 else False,
        ).to(device)
    else:
        raise ValueError(f"{config.reconstruction.method} is not a supported algorithm")

    # constructing algorithm name by appending pre and post process
    algorithm_name = config.reconstruction.method
    if config.reconstruction.pre_process.network is not None:
        algorithm_name = pre_process_name + "_" + algorithm_name
    if config.reconstruction.post_process.network is not None:
        algorithm_name += "_" + post_process_name

    # print number of trainable parameters
    n_param = sum(p.numel() for p in recon.parameters() if p.requires_grad)
    if mask is not None:
        n_param += sum(p.numel() for p in mask.parameters() if p.requires_grad)
    log.info(f"Training model with {n_param} parameters")

    log.info(f"Setup time : {time.time() - start_time} s")
    log.info(f"PSF shape : {psf.shape}")
    log.info(f"Results saved in {save}")
    trainer = Trainer(
        recon=recon,
        train_dataset=train_set,
        test_dataset=test_set,
        mask=mask,
        batch_size=config.training.batch_size,
        eval_batch_size=config.training.eval_batch_size,
        loss=config.loss,
        lpips=config.lpips,
        l1_mask=config.trainable_mask.L1_strength,
        optimizer=config.optimizer,
        skip_NAN=config.training.skip_NAN,
        algorithm_name=algorithm_name,
        metric_for_best_model=config.training.metric_for_best_model,
        save_every=config.training.save_every,
        gamma=config.display.gamma,
        logger=log,
        crop=crop if config.training.crop_preloss else None,
        pre_process=pre_process,
        pre_process_delay=pre_proc_delay,
        pre_process_freeze=config.reconstruction.pre_process.freeze,
        pre_process_unfreeze=config.reconstruction.pre_process.unfreeze,
        post_process=post_process,
        post_process_delay=post_proc_delay,
        post_process_freeze=config.reconstruction.post_process.freeze,
        post_process_unfreeze=config.reconstruction.post_process.unfreeze,
        clip_grad=config.training.clip_grad,
        unrolled_output_factor=config.unrolled_output_factor,
    )

    trainer.train(n_epoch=config.training.epoch, save_pt=save, disp=config.eval_disp_idx)

    log.info(f"Results saved in {save}")


if __name__ == "__main__":
    train_unrolled()



"""

Set DigiCam mask and take picture.

"""

from lensless.hardware.utils import display
import warnings
import hydra
from datetime import datetime
import numpy as np
from slm_controller import slm
from slm_controller.hardware import SLMParam, slm_devices
import matplotlib.pyplot as plt

from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full
from lensless.hardware.aperture import rect_aperture, circ_aperture
from lensless.hardware.utils import set_mask_sensor_distance

import os
from lensless.utils.plot import plot_image, pixel_histogram
from lensless.utils.io import save_image
from lensless.hardware.utils import capture


@hydra.main(version_base=None, config_path="../../configs", config_name="digicam_capture")
def digicam_capture(config):

    rpi_username = config.rpi.username
    rpi_hostname = config.rpi.hostname

    """
    1) Set display image.
    """

    display_config = config.display
    if display_config is not None:
        display(rpi_username=rpi_username, rpi_hostname=rpi_hostname, **display_config)

    """
    2) Set digicam mask.
    """
    digicam_config = config.digicam
    if digicam_config is not None:
        device = digicam_config.device

        shape = slm_devices[device][SLMParam.SLM_SHAPE]
        if not slm_devices[device][SLMParam.MONOCHROME]:
            shape = (3, *shape)
        pixel_pitch = slm_devices[device][SLMParam.PIXEL_PITCH]

        # set mask to sensor distance
        if digicam_config.z is not None:
            set_mask_sensor_distance(digicam_config.z, rpi_username, rpi_hostname)

        center = np.array(digicam_config.center) * pixel_pitch

        # create random pattern
        pattern = None
        if digicam_config.pattern.endswith(".npy"):
            if digicam_config.subpattern:
                assert digicam_config.aperture is not None
                subpattern = np.load(digicam_config.pattern)
                pattern = adafruit_sub2full(subpattern, digicam_config.aperture.center)
            else:
                pattern = np.load(digicam_config.pattern)
        elif digicam_config.pattern == "random":
            rng = np.random.RandomState(config.seed)
            pattern = rng.uniform(low=digicam_config.min_val, high=1, size=shape)
            pattern = (pattern * np.iinfo(np.uint8).max).astype(np.uint8)

        elif digicam_config.pattern == "rect":
            rect_shape = digicam_config.rect_shape
            apert_dim = rect_shape[0] * pixel_pitch[0], rect_shape[1] * pixel_pitch[1]
            ap = rect_aperture(
                apert_dim=apert_dim,
                slm_shape=slm_devices[device][SLMParam.SLM_SHAPE],
                pixel_pitch=pixel_pitch,
                center=center,
            )
            pattern = ap.values
        elif digicam_config.pattern == "circ":
            ap = circ_aperture(
                radius=config.radius * pixel_pitch[0],
                slm_shape=slm_devices[device][SLMParam.SLM_SHAPE],
                pixel_pitch=pixel_pitch,
                center=center,
            )
            pattern = ap.values
        else:
            raise ValueError(f"Pattern {digicam_config.pattern} not supported")

        # apply aperture
        if digicam_config.aperture is not None and digicam_config.aperture.shape is not None:

            apert_dim = np.array(digicam_config.aperture.shape) * np.array(pixel_pitch)
            ap = rect_aperture(
                apert_dim=apert_dim,
                slm_shape=slm_devices[device][SLMParam.SLM_SHAPE],
                pixel_pitch=pixel_pitch,
                center=np.array(digicam_config.aperture.center) * pixel_pitch,
            )
            aperture = ap.values
            aperture[aperture > 0] = 1
            pattern = pattern * aperture

        # save pattern
        if not digicam_config.pattern.endswith(".npy") and config.save:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            pattern_fn = f"{device}_{digicam_config.pattern}_pattern_{timestamp}.npy"
            np.save(pattern_fn, pattern)
            if config.verbose:
                print(f"Saved pattern to {pattern_fn}")

        if config.verbose:
            print("Pattern shape : ", pattern.shape)
            print("Pattern dtype : ", pattern.dtype)
            print("Pattern min   : ", pattern.min())
            print("Pattern max   : ", pattern.max())

        assert pattern is not None

        if config.verbose:
            n_nonzero = np.count_nonzero(pattern)
            print(f"Nonzero pixels: {n_nonzero}")

        set_programmable_mask(pattern, device, rpi_username, rpi_hostname, verbose=config.verbose)

        # preview mask
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            s = slm.create(device)
            s._show_preview(pattern)
            plt.savefig("preview.png")

    """
    3) Take picture with Raspberry Pi camera.
    """

    plot = config.plot
    if config.save:
        if config.output is not None:
            # make sure output directory exists
            os.makedirs(config.output, exist_ok=True)
            save = config.output
        else:
            save = os.getcwd()
    else:
        save = False

    fn = config.capture.raw_data_fn
    localfile, img = capture(
        fn=fn,
        rpi_username=rpi_username,
        rpi_hostname=rpi_hostname,
        verbose=config.verbose,
        output_dir=save,
        **config.capture,
    )

    # save image as viewable 8 bit
    if config.capture.nbits_out != 8:
        fp = os.path.join(save, f"{fn}_8bit.png")
        save_image(img, fp)

    # plot RGB
    if plot:
        plot_image(img, gamma=config.gamma)
        if save:
            plt.savefig(os.path.join(save, f"{fn}_plot.png"))
        pixel_histogram(img)
        if save:
            plt.savefig(os.path.join(save, f"{fn}_hist.png"))

        plt.show()

    if save:
        print(f"\nSaved plots to: {save}")


if __name__ == "__main__":
    digicam_capture()



import numpy as np
from lensless.recon import ReconstructionAlgorithm
from scipy import fft

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class ADMM_PnP(ReconstructionAlgorithm):
    """
    Object for applying ADMM (Alternating Direction Method of Multipliers) with
    a non-negativity constraint and a total variation (TV) prior.

    Paper about ADMM: https://web.stanford.edu/~boyd/papers/pdf/admm_distr_stats.pdf

    Slides about ADMM: https://web.stanford.edu/class/ee364b/lectures/admm_slides.pdf

    """

    def __init__(
        self,
        psf,
        proj,
        dtype=None,
        mu1=1e-6,
        mu2=1e-5,
        mu3=4e-5,
        tau=0.0001,
        pad=False,
        norm="backward",
        use_projection_dual=False,
        **kwargs,
    ):
        """

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        proj : :py:class:`function`
            Projection function to apply at each iteration.
        dtype : float32 or float64
            Data type to use for optimization. Default is float32.
        mu1 : float
            Step size for updating primal/dual variables.
        mu2 : float
            Step size for updating primal/dual variables.
        mu3 : float
            Step size for updating primal/dual variables.
        tau : float
            Weight for L1 norm of `psi` applied to the image estimate.
        pad : bool
            Whether to pad the image with zeros before applying the PSF. Default
            is False, as optimized data is already padded.
        norm : str
            Normalization to use for the convolution. Options are "forward",
            "backward", and "ortho". Default is "backward".
        use_projection_dual : bool
            Whether to use dual update for the user specify projection. This result in an algorithm closer to ADMM.
            During our testing, it seems to give better result with few iteration (<20) but worst otherwise.  Default is False.
        """
        self._mu1 = mu1
        self._mu2 = mu2
        self._mu3 = mu3
        self._tau = tau

        # TODO add check for proj
        self._proj = proj
        self._use_projection_dual = use_projection_dual

        # 3D ADMM is not supported yet
        assert len(psf.shape) == 4, "PSF must be 4D: (depth, height, width, channels)."
        if psf.shape[0] > 1:
            raise NotImplementedError(
                "3D ADMM is not supported yet, use gradient descent or APGD instead."
            )

        # call reset() to initialize matrices
        super(ADMM_PnP, self).__init__(psf, dtype, pad=pad, norm=norm, **kwargs)

        # precompute_R_divmat (self._H computed by constructor with reset())
        if self.is_torch:
            self._R_divmat = 1.0 / (
                self._mu1 * (torch.abs(self._convolver._Hadj * self._convolver._H))
                + self._mu2
                + self._mu3
            ).type(self._complex_dtype)
        else:
            self._R_divmat = 1.0 / (
                self._mu1 * (np.abs(self._convolver._Hadj * self._convolver._H))
                + self._mu2 * np.abs(self._PsiTPsi)
                + self._mu3
            ).astype(self._complex_dtype)

    def reset(self):
        if self.is_torch:
            # TODO initialize without padding
            # initialize image estimate as [Batch, Depth, Height, Width, Channels]
            if self._initial_est is not None:
                self._image_est = self._initial_est
            else:
                self._image_est = torch.zeros([1] + self._padded_shape, dtype=self._dtype).to(
                    self._psf.device
                )

            # self._image_est = torch.zeros_like(self._psf)
            self._X = torch.zeros_like(self._image_est)
            self._U = torch.zeros_like(self._proj(self._image_est))
            self._W = torch.zeros_like(self._X)
            if self._image_est.max():
                # if non-zero
                # self._forward_out = self._forward()
                self._forward_out = self._convolver.convolve(self._image_est)
            else:
                self._forward_out = torch.zeros_like(self._X)

            self._xi = torch.zeros_like(self._image_est)
            self._eta = torch.zeros_like(self._U)
            self._rho = torch.zeros_like(self._X)

            # precompute_X_divmat
            self._X_divmat = 1.0 / (self._convolver._pad(torch.ones_like(self._psf)) + self._mu1)
            # self._X_divmat = 1.0 / (torch.ones_like(self._psf) + self._mu1)

        else:
            if self._initial_est is not None:
                self._image_est = self._initial_est
            else:
                self._image_est = np.zeros([1] + self._padded_shape, dtype=self._dtype)

            # self._U = np.zeros(np.r_[self._padded_shape, [2]], dtype=self._dtype)
            self._X = np.zeros_like(self._image_est)
            self._U = np.zeros_like(self._proj(self._image_est))
            self._W = np.zeros_like(self._X)
            if self._image_est.max():
                # if non-zero
                # self._forward_out = self._forward()
                self._forward_out = self._convolver.convolve(self._image_est)
            else:
                self._forward_out = np.zeros_like(self._X)

            self._xi = np.zeros_like(self._image_est)
            self._eta = np.zeros_like(self._U)
            self._rho = np.zeros_like(self._X)

            # precompute_X_divmat
            self._X_divmat = 1.0 / (
                self._convolver._pad(np.ones(self._psf_shape, dtype=self._dtype)) + self._mu1
            )

    def _U_update(self):
        """Total variation update."""
        # to avoid computing sparse operator twice
        if self._use_projection_dual:
            self._U = self._proj(self._U + self._eta / self._mu2)
        else:
            self._U = self._proj(self._image_est)

    def _X_update(self):
        # to avoid computing forward model twice
        # self._X = self._X_divmat * (self._xi + self._mu1 * self._forward_out + self._data)
        self._X = self._X_divmat * (
            self._xi + self._mu1 * self._forward_out + self._convolver._pad(self._data)
        )

    def _W_update(self):
        """Non-negativity update"""
        if self.is_torch:
            self._W = torch.maximum(
                self._rho / self._mu3 + self._image_est, torch.zeros_like(self._image_est)
            )
        else:
            self._W = np.maximum(self._rho / self._mu3 + self._image_est, 0)

    def _image_update(self):
        rk = (
            (self._mu3 * self._W - self._rho)
            + self._mu2 * self._U
            + self._convolver.deconvolve(self._mu1 * self._X - self._xi)
        )

        # rk = self._convolver._pad(rk)

        if self.is_torch:
            freq_space_result = self._R_divmat * torch.fft.rfft2(rk, dim=(-3, -2))
            self._image_est = torch.fft.irfft2(freq_space_result, dim=(-3, -2))
        else:
            freq_space_result = self._R_divmat * fft.rfft2(rk, axes=(-3, -2))
            self._image_est = fft.irfft2(freq_space_result, axes=(-3, -2))

        # self._image_est = self._convolver._crop(res)

    def _xi_update(self):
        # to avoid computing forward model twice
        self._xi += self._mu1 * (self._forward_out - self._X)

    def _eta_update(self):
        # to avoid finite difference operataion again?
        self._eta += self._mu2 * (self._image_est - self._U)

    def _rho_update(self):
        self._rho += self._mu3 * (self._image_est - self._W)

    def _update(self, iter):
        self._U_update()
        if torch.isnan(self._U).any():
            raise RuntimeError("NaN encountered in U update.")
        self._X_update()
        if torch.isnan(self._X).any():
            raise RuntimeError("NaN encountered in X update.")
        self._W_update()
        if torch.isnan(self._W).any():
            raise RuntimeError("NaN encountered in W update.")
        self._image_update()
        if torch.isnan(self._image_est).any():
            raise RuntimeError("NaN encountered in image update.")

        # update forward and sparse operators
        self._forward_out = self._convolver.convolve(self._image_est)

        self._xi_update()
        if self._use_projection_dual:
            self._eta_update()
        self._rho_update()

    def _form_image(self):
        image = self._convolver._crop(self._image_est)

        # # TODO without cropping
        # image = self._image_est

        image[image < 0] = 0
        return image


import os
import glob
import numpy as np
from PIL import Image

def save_u16_to_tiff(u16in, size, tiff_filename):
    """
    Since Pillow has poor support for 16-bit TIFF, we make our own
    save function to properly save a 16-bit TIFF.
    """
    # write 16-bit TIFF image

    # PIL interprets mode 'I;16' as "uint16, little-endian"
    img_out = Image.new('I;16', size)

    if HAS_NUMPY:
        # make sure u16in little-endian, output bytes
        outpil = u16in.astype(u16in.dtype.newbyteorder("<")).tobytes()
    else:
        # little-endian u16 format
        outpil = struct.pack(
                "<%dH"%(len(u16in)),
                *u16in
                )
    img_out.frombytes(outpil)
    img_out.save(tiff_filename)


dir_diffuser = "/scratch/bezzam/DiffuserCam_mirflickr/dataset/diffuser_images"
files_diffuser = glob.glob(os.path.join(dir_diffuser, "*.npy"))

# load a single file
diffuser_np = np.load(files_diffuser[0])

# conver to 16-bit
diffuser_np = (diffuser_np * 65535).astype(np.uint16)

# import pudb; pudb.set_trace()  # XXX BREAKPOINT

# save as 16-bit image

img = Image.fromarray(diffuser_np)
img.save("diffuser.tiff")

# load the 16-bit image
img = Image.open("diffuser.tiff")
diffuser_img = np.array(img)

import pudb; pudb.set_trace()  # XXX BREAKPOINT


import torchvision.datasets as dset
import numpy as np
import time
import os
import subprocess
import pathlib as plib
import click


@click.command()
@click.option(
    "--hostname",
    type=str,
    help="Hostname or IP address for display device.",
)
@click.option(
    "--camera_hostname",
    type=str,
    help="Hostname or IP address for capture device. If not provided, same as hostname",
)
@click.option(
    "--output_dir",
    type=str,
    help="Output directory for measured images.",
)
@click.option(
    "--n_files",
    type=int,
    help="Number of files to collect. Default is all files.",
)
@click.option(
    "--test",
    is_flag=True,
    help="Measure test set, otherwise do train.",
)
@click.option(
    "--bayer",
    is_flag=True,
    help="Copy over raw Bayer data. Otherwise do de-mosaicing and downsampling on Raspberry Pi.",
)
@click.option(
    "--downsample",
    type=float,
    default=128,
    help="Amount to downsample.",
)
@click.option("-v", "--verbose", is_flag=True)
def collect_mnist(hostname, camera_hostname, output_dir, n_files, verbose, test, bayer, downsample):
    assert hostname is not None
    assert output_dir is not None

    # TODO : save metadata in JSON
    progress = 10

    # display param
    display_python = "~/LenslessPiCam/lensless_env/bin/python"
    display_tmp_file = "~/tmp_display.png"
    display_image_prep_script = "~/LenslessPiCam/scripts/measure/prep_display_image.py"
    display_image_path = "~/LenslessPiCam_display/test.png"
    display_res = np.array((1920, 1200))
    hshift = 0
    vshift = 0
    pad = 50
    brightness = 100
    local_display_file = "display_tmp.png"

    # camera param
    camera_python = "~/LenslessPiCam/lensless_env/bin/python"
    if camera_hostname is None:
        camera_hostname = hostname
    sensor_mode = "0"
    exp = 0.2
    config_pause = 0
    capture_script = "~/LenslessPiCam/scripts/measure/on_device_capture.py"

    # load datasets
    root = "./data"
    tr = None
    # tr = transforms.ToTensor()
    dataset = dset.MNIST(root=root, train=not test, download=True, transform=tr)

    print("\nNumber of files :", len(dataset))
    if n_files:
        print(f"TEST : collecting first {n_files} files!")
    else:
        n_files = len(dataset)

    # make output directory
    output_dir = plib.Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    # loop over files
    if test:
        subdir = output_dir / "test"
    else:
        subdir = output_dir / "train"
    subdir.mkdir(exist_ok=True)
    labels = []
    start_time = time.time()
    for i in range(n_files):
        if verbose:
            print(f"\nFILE : {i+1} / {n_files}")
        img, label = dataset[i]
        labels.append(label)

        # TODO check if measurement already exists
        output_fp = subdir / f"img{i}.png"
        if not os.path.isfile(output_fp):
            # load and save as PNG
            img.save(local_display_file)

            # send to display
            if verbose:
                print("-- Displaying picture...")
            # -- copy to RPi
            os.system('scp %s "pi@%s:%s" ' % (local_display_file, hostname, display_tmp_file))

            # -- prep image on Pi
            prep_command = f"{display_python} {display_image_prep_script} --fp {display_tmp_file} \
                --pad {pad} --vshift {vshift} --hshift {hshift} --screen_res {display_res[0]} {display_res[1]} \
                --brightness {brightness} --output_path {display_image_path} "
            if verbose:
                print(f"Command : {prep_command}")
            subprocess.Popen(
                ["ssh", "pi@%s" % hostname, prep_command],
                shell=False,
            )
            # time.sleep(2)

            # take picture
            remote_fn = "remote_capture"
            pic_command = (
                f"{camera_python} {capture_script} --fn {remote_fn} --exp {exp} --iso 100 "
                f"--config_pause {config_pause} --sensor_mode {sensor_mode} --nbits_out 8 --sixteen --legacy"
            )
            if not bayer:
                pic_command += " --gray"
                if downsample:
                    pic_command += f" --down {downsample}"
            if verbose:
                print("-- Taking picture...")
                print(f"Command : {pic_command}")
            ssh = subprocess.Popen(
                ["ssh", "pi@%s" % camera_hostname, pic_command],
                shell=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            result = ssh.stdout.readlines()
            error = ssh.stderr.readlines()
            if error != []:
                raise ValueError("CAPTURE ERROR: %s" % error)
            if result == []:
                error = ssh.stderr.readlines()
                raise ValueError("CAPTURE ERROR: %s" % error)

            # copy over file
            remotefile = f"~/{remote_fn}.png"
            if verbose:
                print(f"Copying over picture as {output_fp}...")
            os.system('scp "pi@%s:%s" %s' % (camera_hostname, remotefile, output_fp))

        if (i + 1) % progress == 0:
            proc_time = time.time() - start_time
            print(f"\n{i+1} / {n_files}, {proc_time / 60.:.3f} minutes")

    with open(subdir / "labels.txt", "w") as f:
        for item in labels:
            f.write("%s\n" % item)

    proc_time = time.time() - start_time
    print(f"Finished, {proc_time/60.:.3f} minutes.")

    if os.path.isfile(local_display_file):
        os.remove(local_display_file)


if __name__ == "__main__":
    collect_mnist()



# #############################################################################
# utils.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


import json
import math
import numpy as np
import matplotlib.pyplot as plt
import time
from hydra.utils import get_original_cwd
import os
import torch
from lensless.eval.benchmark import benchmark
from lensless.hardware.trainable_mask import TrainableMask
from tqdm import tqdm
from lensless.recon.drunet.network_unet import UNetRes
from lensless.utils.io import save_image
from lensless.utils.plot import plot_image


def load_drunet(model_path=None, n_channels=3, requires_grad=False):
    """
    Load a pre-trained Drunet model.

    Parameters
    ----------
    model_path : str, optional
        Path to pre-trained model. Download if not provided.
    n_channels : int
        Number of channels in input image.
    requires_grad : bool
        Whether to require gradients for model parameters.

    Returns
    -------
    model : :py:class:`torch.nn.Module`
        Loaded model.
    """

    if model_path is None:
        this_file_path = os.path.dirname(os.path.realpath(__file__))
        model_path = os.path.join(this_file_path, "..", "..", "models", "drunet_color.pth")
        if not os.path.exists(model_path):
            try:
                from torchvision.datasets.utils import download_url
            except ImportError:
                exit()
            msg = "Do you want to download the pretrained DRUNet model (130MB)?"

            # default to yes if no input is given
            valid = input("%s (Y/n) " % msg).lower() != "n"
            output_path = os.path.join(get_original_cwd(), "models")
            if valid:
                url = "https://drive.switch.ch/index.php/s/jTdeMHom025RFRQ/download"
                filename = "drunet_color.pth"
                download_url(url, output_path, filename=filename)

    assert os.path.exists(model_path), f"Model path {model_path} does not exist"

    model = UNetRes(
        in_nc=n_channels + 1,
        out_nc=n_channels,
        nc=[64, 128, 256, 512],
        nb=4,
        act_mode="R",
        downsample_mode="strideconv",
        upsample_mode="convtranspose",
    )
    model.load_state_dict(torch.load(model_path), strict=True)
    model.eval()
    for _, v in model.named_parameters():
        v.requires_grad = requires_grad

    return model


def apply_denoiser(model, image, noise_level=10, device="cpu", mode="inference"):
    """
    Apply a pre-trained denoising model with input in the format Channel, Height, Width.
    An additionnal channel is added for the noise level as done in Drunet.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        Drunet compatible model. Its input must consist of 4 channels (RGB + noise level) and output an RGB image both in CHW format.
    image : :py:class:`torch.Tensor`
        Input image.
    noise_level : float or :py:class:`torch.Tensor`
        Noise level in the image within [0, 255].
    device : str
        Device to use for computation. Can be "cpu" or "cuda".
    mode : str
        Mode to use for model. Can be "inference" or "train".

    Returns
    -------
    image : :py:class:`torch.Tensor`
        Reconstructed image.
    """
    assert noise_level > 0
    assert noise_level <= 255

    # convert from NDHWC to NCHW
    depth = image.shape[-4]
    image = image.movedim(-1, -3)
    image = image.reshape(-1, *image.shape[-3:])
    # pad image H and W to next multiple of 8
    top = (8 - image.shape[-2] % 8) // 2
    bottom = (8 - image.shape[-2] % 8) - top
    left = (8 - image.shape[-1] % 8) // 2
    right = (8 - image.shape[-1] % 8) - left
    image = torch.nn.functional.pad(image, (left, right, top, bottom), mode="constant", value=0)
    # add noise level as extra channel
    image = image.to(device)
    if isinstance(noise_level, torch.Tensor):
        noise_level = noise_level / 255.0
    else:
        noise_level = torch.tensor([noise_level / 255.0]).to(device)

    image = torch.cat(
        (
            image,
            noise_level.repeat(image.shape[0], 1, image.shape[2], image.shape[3]),
        ),
        dim=1,
    )

    # apply model
    if mode == "inference":
        with torch.no_grad():
            image = model(image)
    elif mode == "train":
        image = model(image)
    else:
        raise ValueError("mode must be 'inference' or 'train'")

    # remove padding
    image = image[:, :, top:-bottom, left:-right]
    # convert back to NDHWC
    image = image.movedim(-3, -1)
    image = image.reshape(-1, depth, *image.shape[-3:])
    return image


def get_drunet_function(model, device="cpu", mode="inference"):
    """
    Return a processing function that applies the DruNet model to an image.
    Legacy function to work with pre-trained models, use get_drunet_function_v2 instead.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        DruNet like denoiser model
    device : str
        Device to use for computation. Can be "cpu" or "cuda".
    mode : str
        Mode to use for model. Can be "inference" or "train".
    """

    def process(image, noise_level):
        x_max = torch.amax(image, dim=(-2, -3), keepdim=True) + 1e-6
        image = apply_denoiser(
            model,
            image,
            noise_level=noise_level,
            device=device,
            mode=mode,
        )
        image = torch.clip(image, min=0.0) * x_max
        return image

    return process


def get_drunet_function_v2(model, device="cpu", mode="inference"):
    """
    Return a processing function that applies the DruNet model to an image.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        DruNet like denoiser model
    device : str
        Device to use for computation. Can be "cpu" or "cuda".
    mode : str
        Mode to use for model. Can be "inference" or "train".
    """

    def process(image, noise_level):
        x_max = torch.amax(image, dim=(-1, -2, -3, -4), keepdim=True) + 1e-6
        image = apply_denoiser(
            model,
            image / x_max,
            noise_level=noise_level,
            device=device,
            mode=mode,
        )
        image = torch.clip(image, min=0.0) * x_max
        return image

    return process


def measure_gradient(model):
    """
    Helper function to measure L2 norm of the gradient of a model.

    Parameters
    ----------
    model : :py:class:`torch.nn.Module`
        Model to measure gradient of.

    Returns
    -------
    Float
        L2 norm of the gradient of the model.
    """
    total_norm = 0.0
    for p in model.parameters():
        param_norm = p.grad.detach().data.norm(2)
        total_norm += param_norm.item() ** 2
    total_norm = total_norm**0.5
    return total_norm


def create_process_network(network, depth=4, device="cpu", nc=None):
    """
    Helper function to create a process network.

    Parameters
    ----------
    network : str
        Name of network to use. Can be "DruNet" or "UnetRes".
    depth : int
        Depth of network.
    device : str
        Device to use for computation. Can be "cpu" or "cuda". Defaults to "cpu".

    Returns
    -------
    :py:class:`torch.nn.Module`
        New process network. Already trained for Drunet.
    """

    if nc is None:
        nc = [64, 128, 256, 512]
    else:
        assert len(nc) == 4

    if network == "DruNet":
        from lensless.recon.utils import load_drunet

        process = load_drunet(requires_grad=True).to(device)
        process_name = "DruNet"
    elif network == "UnetRes":
        from lensless.recon.drunet.network_unet import UNetRes

        n_channels = 3
        process = UNetRes(
            in_nc=n_channels + 1,
            out_nc=n_channels,
            nc=nc,
            nb=depth,
            act_mode="R",
            downsample_mode="strideconv",
            upsample_mode="convtranspose",
        ).to(device)
        process_name = "UnetRes_d" + str(depth)
    else:
        process = None
        process_name = None

    return (process, process_name)


class Trainer:
    def __init__(
        self,
        recon,
        train_dataset,
        test_dataset,
        test_size=0.15,
        mask=None,
        batch_size=4,
        eval_batch_size=10,
        loss="l2",
        lpips=None,
        l1_mask=None,
        optimizer=None,
        skip_NAN=False,
        algorithm_name="Unknown",
        metric_for_best_model=None,
        save_every=None,
        gamma=None,
        logger=None,
        crop=None,
        clip_grad=1.0,
        unrolled_output_factor=False,
        # for adding components during training
        pre_process=None,
        pre_process_delay=None,
        pre_process_freeze=None,
        pre_process_unfreeze=None,
        post_process=None,
        post_process_delay=None,
        post_process_freeze=None,
        post_process_unfreeze=None,
    ):
        """
        Class to train a reconstruction algorithm. Inspired by Trainer from `HuggingFace <https://huggingface.co/docs/transformers/main_classes/trainer>`__.

        The train and test metrics at the end of each epoch can be found in ``self.metrics``,
        with "LOSS" being the train loss. The test loss can be found in "MSE" (if loss is "l2") or
        "MAE" (if loss is "l1"). If ``lpips`` is not None, the LPIPS loss is also added
        to the train loss, such that the test loss can be computed as "MSE" + ``lpips`` * "LPIPS_Vgg"
        (or "MAE" + ``lpips`` * "LPIPS_Vgg").

        Parameters
        ----------
        recon : :py:class:`lensless.TrainableReconstructionAlgorithm`
            Reconstruction algorithm to train.
        train_dataset : :py:class:`torch.utils.data.Dataset`
            Dataset to use for training.
        test_dataset : :py:class:`torch.utils.data.Dataset`
            Dataset to use for testing.
        test_size : float, optional
            If test_dataset is None, fraction of the train dataset to use for testing, by default 0.15.
        mask : TrainableMask, optional
            Trainable mask to use for training. If none, training with fix psf, by default None.
        batch_size : int, optional
            Batch size to use for training, by default 4.
        eval_batch_size : int, optional
            Batch size to use for evaluation, by default 10.
        loss : str, optional
            Loss function to use for training "l1" or "l2", by default "l2".
        lpips : float, optional
            the weight of the lpips(VGG) in the total loss. If None ignore. By default None.
        l1_mask : float, optional
            the weight of the l1 norm of the mask in the total loss. If None ignore. By default None.
        optimizer : dict
            Optimizer configuration.
        skip_NAN : bool, optional
            Whether to skip update if any gradiant are NAN (True) or to throw an error(False), by default False
        algorithm_name : str, optional
            Algorithm name for logging, by default "Unknown".
        metric_for_best_model : str, optional
            Metric to use for saving the best model. If None, will default to evaluation loss. Default is None.
        save_every : int, optional
            Save model every ``save_every`` epochs. If None, just save best model.
        gamma : float, optional
            Gamma correction to apply to PSFs when plotting. If None, no gamma correction is applied. Default is None.
        logger : :py:class:`logging.Logger`, optional
            Logger to use for logging. If None, just print to terminal. Default is None.
        crop : dict, optional
            Crop to apply to images before computing loss (by applying a mask). If None, no crop is applied. Default is None.
        unrolled_output_factor : float, optional
            How much of the unrolled loss to add to the total loss. If False, no unrolled loss is added. Default is False. Only applicable if a post-processor is used.
        pre_process : :py:class:`torch.nn.Module`, optional
            Pre process component to add during training. Default is None.
        pre_process_delay : int, optional
            Epoch at which to add pre process component. Default is None.
        pre_process_freeze : int, optional
            Epoch at which to freeze pre process component. Default is None.
        pre_process_unfreeze : int, optional
            Epoch at which to unfreeze pre process component. Default is None.
        post_process : :py:class:`torch.nn.Module`, optional
            Post process component to add during training. Default is None.
        post_process_delay : int, optional
            Epoch at which to add post process component. Default is None.
        post_process_freeze : int, optional
            Epoch at which to freeze post process component. Default is None.
        post_process_unfreeze : int, optional
            Epoch at which to unfreeze post process component. Default is None.


        """
        global print

        self.device = recon._psf.device
        self.logger = logger
        if self.logger is not None:
            self.print = self.logger.info
        else:
            self.print = print
        self.recon = recon

        self.pre_process = pre_process
        self.pre_process_delay = pre_process_delay
        self.pre_process_freeze = pre_process_freeze
        self.pre_process_unfreeze = pre_process_unfreeze
        self.pre_process_delay = pre_process_delay
        if pre_process_delay is not None:
            assert pre_process is not None

        self.post_process = post_process
        self.post_process_delay = post_process_delay
        self.post_process_freeze = post_process_freeze
        self.post_process_unfreeze = post_process_unfreeze
        self.post_process_delay = post_process_delay
        if post_process_delay is not None:
            assert post_process is not None

        assert train_dataset is not None
        if test_dataset is None:
            assert test_size < 1.0 and test_size > 0.0
            # split train dataset
            train_size = int((1 - test_size) * len(train_dataset))
            test_size = len(train_dataset) - train_size
            train_dataset, test_dataset = torch.utils.data.random_split(
                train_dataset, [train_size, test_size]
            )
            self.print(f"Train size : {train_size}, Test size : {test_size}")

        self.train_dataloader = torch.utils.data.DataLoader(
            dataset=train_dataset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=(self.device != "cpu"),
        )
        self.test_dataset = test_dataset
        self.lpips = lpips
        self.skip_NAN = skip_NAN
        self.eval_batch_size = eval_batch_size

        self.mask = mask
        if mask is not None:
            assert isinstance(mask, TrainableMask)
            self.use_mask = True
        else:
            self.use_mask = False

        self.l1_mask = l1_mask
        self.gamma = gamma

        # loss
        if loss == "l2":
            self.Loss = torch.nn.MSELoss()
        elif loss == "l1":
            self.Loss = torch.nn.L1Loss()
        else:
            raise ValueError(f"Unsuported loss : {loss}")

        # -- Lpips loss
        if lpips:
            try:
                import lpips

                self.Loss_lpips = lpips.LPIPS(net="vgg").to(self.device)
            except ImportError:
                return ImportError(
                    "lpips package is need for LPIPS loss. Install using : pip install lpips"
                )

        self.crop = crop

        # -- adding unrolled loss
        self.unrolled_output_factor = unrolled_output_factor
        if self.unrolled_output_factor:
            assert self.unrolled_output_factor > 0
            assert self.post_process is not None
            assert self.post_process_delay is None
            assert self.post_process_unfreeze is None
            assert self.post_process_freeze is None

        # optimizer
        self.clip_grad_norm = clip_grad
        self.optimizer_config = optimizer
        self.set_optimizer()

        self.metrics = {
            "LOSS": [],  # train loss
            "LOSS_TEST": [],  # test loss
            "MSE": [],
            "MAE": [],
            "LPIPS_Vgg": [],
            "LPIPS_Alex": [],
            "PSNR": [],
            "SSIM": [],
            "ReconstructionError": [],
            "n_iter": self.recon._n_iter,
            "algorithm": algorithm_name,
            "metric_for_best_model": metric_for_best_model,
            "best_epoch": 0,
            "best_eval_score": 0
            if metric_for_best_model == "PSNR" or metric_for_best_model == "SSIM"
            else np.inf,
        }
        if self.unrolled_output_factor:
            # -- add unrolled metrics
            for key in ["MSE", "MAE", "LPIPS_Vgg", "LPIPS_Alex", "PSNR", "SSIM"]:
                self.metrics[key + "_unrolled"] = []
        if metric_for_best_model is not None:
            assert metric_for_best_model in self.metrics.keys()
        self.save_every = save_every

        # Backward hook that detect NAN in the gradient and print the layer weights
        if not self.skip_NAN:

            def detect_nan(grad):
                if torch.isnan(grad).any():
                    if self.logger:
                        self.logger.info(grad)
                    else:
                        print(grad, flush=True)
                    for name, param in recon.named_parameters():
                        if param.requires_grad:
                            self.print(name, param)
                    raise ValueError("Gradient is NaN")
                return grad

            for param in recon.parameters():
                if param.requires_grad:
                    param.register_hook(detect_nan)
                    if param.requires_grad:
                        param.register_hook(detect_nan)

    def set_optimizer(self, last_epoch=-1):

        if self.optimizer_config.type == "Adam":
            parameters = [{"params": self.recon.parameters()}]
            self.optimizer = torch.optim.Adam(parameters, lr=self.optimizer_config.lr)
        else:
            raise ValueError(f"Unsupported optimizer : {self.optimizer_config.type}")

        # Scheduler
        if self.optimizer_config.slow_start:

            def learning_rate_function(epoch):
                if epoch == 0:
                    return self.optimizer_config.slow_start
                elif epoch == 1:
                    return math.sqrt(self.optimizer_config.slow_start)
                else:
                    return 1

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer, lr_lambda=learning_rate_function, last_epoch=last_epoch
            )

        elif self.optimizer_config.step:

            self.scheduler = torch.optim.lr_scheduler.StepLR(
                self.optimizer,
                step_size=self.optimizer_config.step,
                gamma=self.optimizer_config.gamma,
                last_epoch=last_epoch,
                verbose=True,
            )

        else:

            def learning_rate_function(epoch):
                return 1

            self.scheduler = torch.optim.lr_scheduler.LambdaLR(
                self.optimizer, lr_lambda=learning_rate_function, last_epoch=last_epoch
            )

    def train_epoch(self, data_loader):
        """
        Train for one epoch.

        Parameters
        ----------
        data_loader : :py:class:`torch.utils.data.DataLoader`
            Data loader to use for training.

        Returns
        -------
        float
            Mean loss of the epoch.
        """
        mean_loss = 0.0
        i = 1.0
        pbar = tqdm(data_loader)
        for X, y in pbar:
            # send to device
            X = X.to(self.device)
            y = y.to(self.device)

            # update psf according to mask
            if self.use_mask:
                self.recon._set_psf(self.mask.get_psf().to(self.device))

            # forward pass
            y_pred = self.recon.batch_call(X.to(self.device))
            if self.unrolled_output_factor:
                unrolled_out = y_pred[1]
                y_pred = y_pred[0]

            # normalizing each output
            eps = 1e-12
            y_pred_max = torch.amax(y_pred, dim=(-1, -2, -3), keepdim=True) + eps
            y_pred = y_pred / y_pred_max

            # normalizing y
            y_max = torch.amax(y, dim=(-1, -2, -3), keepdim=True) + eps
            y = y / y_max

            self.optimizer.zero_grad(set_to_none=True)
            # convert to CHW for loss and remove depth
            y_pred = y_pred.reshape(-1, *y_pred.shape[-3:]).movedim(-1, -3)
            y = y.reshape(-1, *y.shape[-3:]).movedim(-1, -3)

            # extraction region of interest for loss
            if self.crop is not None:
                y_pred = y_pred[
                    ...,
                    self.crop["vertical"][0] : self.crop["vertical"][1],
                    self.crop["horizontal"][0] : self.crop["horizontal"][1],
                ]
                y = y[
                    ...,
                    self.crop["vertical"][0] : self.crop["vertical"][1],
                    self.crop["horizontal"][0] : self.crop["horizontal"][1],
                ]

            loss_v = self.Loss(y_pred, y)

            # add LPIPS loss
            if self.lpips:

                if y_pred.shape[1] == 1:
                    # if only one channel, repeat for LPIPS
                    y_pred = y_pred.repeat(1, 3, 1, 1)
                    y = y.repeat(1, 3, 1, 1)

                # value for LPIPS needs to be in range [-1, 1]
                loss_v = loss_v + self.lpips * torch.mean(
                    self.Loss_lpips(2 * y_pred - 1, 2 * y - 1)
                )
            if self.use_mask and self.l1_mask:
                loss_v = loss_v + self.l1_mask * torch.mean(torch.abs(self.mask._mask))

            if self.unrolled_output_factor:
                # -- normalize
                unrolled_out_max = torch.amax(unrolled_out, dim=(-1, -2, -3), keepdim=True) + eps
                unrolled_out = unrolled_out / unrolled_out_max

                # -- convert to CHW for loss and remove depth
                unrolled_out = unrolled_out.reshape(-1, *unrolled_out.shape[-3:]).movedim(-1, -3)

                # -- extraction region of interest for loss
                if self.crop is not None:
                    unrolled_out = unrolled_out[
                        ...,
                        self.crop["vertical"][0] : self.crop["vertical"][1],
                        self.crop["horizontal"][0] : self.crop["horizontal"][1],
                    ]

                # -- compute unrolled output loss
                loss_unrolled = self.Loss(unrolled_out, y)

                # -- add LPIPS loss
                if self.lpips:
                    if unrolled_out.shape[1] == 1:
                        # if only one channel, repeat for LPIPS
                        unrolled_out = unrolled_out.repeat(1, 3, 1, 1)

                    # value for LPIPS needs to be in range [-1, 1]
                    loss_unrolled = loss_unrolled + self.lpips * torch.mean(
                        self.Loss_lpips(2 * unrolled_out - 1, 2 * y - 1)
                    )

                # -- add unrolled loss to total loss
                loss_v = loss_v + self.unrolled_output_factor * loss_unrolled

            # backward pass
            loss_v.backward()

            if self.clip_grad_norm is not None:
                torch.nn.utils.clip_grad_norm_(self.recon.parameters(), self.clip_grad_norm)

            # if any gradient is NaN, skip training step
            if self.skip_NAN:
                is_NAN = False
                for param in self.recon.parameters():
                    if param.grad is not None and torch.isnan(param.grad).any():
                        is_NAN = True
                        break
                if is_NAN:
                    self.print("NAN detected in gradiant, skipping training step")
                    i += 1
                    continue
            self.optimizer.step()

            # update mask
            if self.use_mask:
                self.mask.update_mask()

            mean_loss += (loss_v.item() - mean_loss) * (1 / i)
            pbar.set_description(f"loss : {mean_loss}")
            i += 1

        self.print(f"loss : {mean_loss}")

        return mean_loss

    def evaluate(self, mean_loss, save_pt, epoch, disp=None):
        """
        Evaluate the reconstruction algorithm on the test dataset.

        Parameters
        ----------
        mean_loss : float
            Mean loss of the last epoch.
        save_pt : str
            Path to save metrics dictionary to. If None, no logging of metrics.
        disp : list of int, optional
            Test set examples to visualize at the end of each epoch, by default None.
        """
        if self.test_dataset is None:
            return

        output_dir = None
        if disp is not None:
            output_dir = os.path.join("eval_recon")
            if not os.path.exists(output_dir):
                os.mkdir(output_dir)
            output_dir = os.path.join(output_dir, str(epoch))

        # benchmarking
        current_metrics = benchmark(
            self.recon,
            self.test_dataset,
            batchsize=self.eval_batch_size,
            save_idx=disp,
            output_dir=output_dir,
            crop=self.crop,
            unrolled_output_factor=self.unrolled_output_factor,
        )

        # update metrics with current metrics
        self.metrics["LOSS"].append(mean_loss)
        for key in current_metrics:
            self.metrics[key].append(current_metrics[key])

        if save_pt:
            # save dictionary metrics to file with json
            with open(os.path.join(save_pt, "metrics.json"), "w") as f:
                json.dump(self.metrics, f, indent=4)

        # check best metric
        if self.metrics["metric_for_best_model"] is None:
            eval_loss = current_metrics["MSE"]
            if self.lpips is not None:
                eval_loss += self.lpips * current_metrics["LPIPS_Vgg"]
            if self.use_mask and self.l1_mask:
                eval_loss += self.l1_mask * np.mean(np.abs(self.mask._mask.cpu().detach().numpy()))
            if self.unrolled_output_factor:
                unrolled_loss = current_metrics["MSE_unrolled"]
                if self.lpips is not None:
                    unrolled_loss += self.lpips * current_metrics["LPIPS_Vgg_unrolled"]
                eval_loss += self.unrolled_output_factor * unrolled_loss
        else:
            eval_loss = current_metrics[self.metrics["metric_for_best_model"]]

        self.metrics["LOSS_TEST"].append(eval_loss)
        return eval_loss

    def on_epoch_end(self, mean_loss, save_pt, epoch, disp=None):
        """
        Called at the end of each epoch.

        Parameters
        ----------
        mean_loss : float
            Mean loss of the last epoch.
        save_pt : str
            Path to save metrics dictionary to. If None, no logging of metrics.
        epoch : int
            Current epoch.
        disp : list of int, optional
            Test set examples to visualize at the end of each epoch, by default None.
        """
        if save_pt is None:
            # Use current directory
            save_pt = os.getcwd()

        # save model
        # self.save(path=save_pt, include_optimizer=False)
        epoch_eval_metric = self.evaluate(mean_loss, save_pt, epoch, disp=disp)
        new_best = False
        if (
            self.metrics["metric_for_best_model"] == "PSNR"
            or self.metrics["metric_for_best_model"] == "SSIM"
        ):
            if epoch_eval_metric > self.metrics["best_eval_score"]:
                self.metrics["best_eval_score"] = epoch_eval_metric
                new_best = True
        else:
            if epoch_eval_metric < self.metrics["best_eval_score"]:
                self.metrics["best_eval_score"] = epoch_eval_metric
                new_best = True

        if new_best:
            self.metrics["best_epoch"] = epoch
            self.save(path=save_pt, include_optimizer=False, epoch="BEST")

        if self.save_every is not None and epoch % self.save_every == 0:
            self.save(path=save_pt, include_optimizer=False, epoch=epoch)

    def train(self, n_epoch=1, save_pt=None, disp=None):
        """
        Train the reconstruction algorithm.

        Parameters
        ----------
        n_epoch : int, optional
            Number of epochs to train for, by default 1
        save_pt : str, optional
            Path to save metrics dictionary to. If None, use current directory, by default None
        disp : list of int, optional
            test set examples to visualize at the end of each epoch, by default None.
        """

        start_time = time.time()

        self.evaluate(-1, save_pt, epoch=0, disp=disp)
        for epoch in range(n_epoch):

            # add extra components (if specified)
            changing_n_param = False
            if epoch == self.pre_process_delay:
                self.print("Adding pre process component")
                self.recon.set_pre_process(self.pre_process)
                changing_n_param = True
            if epoch == self.post_process_delay:
                self.print("Adding post process component")
                self.recon.set_post_process(self.post_process)
                changing_n_param = True
            if epoch == self.pre_process_freeze:
                self.print("Freezing pre process")
                self.recon.freeze_pre_process()
                changing_n_param = True
            if epoch == self.post_process_freeze:
                self.print("Freezing post process")
                self.recon.freeze_post_process()
                changing_n_param = True
            if epoch == self.pre_process_unfreeze:
                self.print("Unfreezing pre process")
                self.recon.unfreeze_pre_process()
                changing_n_param = True
            if epoch == self.post_process_unfreeze:
                self.print("Unfreezing post process")
                self.recon.unfreeze_post_process()
                changing_n_param = True

            # count number of parameters with requires_grad = True
            if changing_n_param:
                n_param = sum(p.numel() for p in self.recon.parameters() if p.requires_grad)
                if self.mask is not None:
                    n_param += sum(p.numel() for p in self.mask.parameters() if p.requires_grad)
                self.print(f"Training {n_param} parameters")

            self.print(f"Epoch {epoch} with learning rate {self.scheduler.get_last_lr()}")
            mean_loss = self.train_epoch(self.train_dataloader)
            # offset because of evaluate before loop
            self.on_epoch_end(mean_loss, save_pt, epoch + 1, disp=disp)
            self.scheduler.step()

        self.print(f"Train time [hour] : {(time.time() - start_time) / 3600} h")

    def save(self, epoch, path="recon", include_optimizer=False):
        # create directory if it does not exist
        if not os.path.exists(path):
            os.makedirs(path)
        # save mask
        if self.use_mask:
            # torch.save(self.mask._mask, os.path.join(path, f"mask_epoch{epoch}.pt"))

            # save mask as numpy array
            if self.mask.train_mask_vals:
                np.save(
                    os.path.join(path, f"mask_epoch{epoch}.npy"),
                    self.mask._mask.cpu().detach().numpy(),
                )

            # if color_filter is an attribute
            if hasattr(self.mask, "color_filter") and self.mask.color_filter is not None:
                # save save numpy array
                np.save(
                    os.path.join(path, f"mask_color_filter_epoch{epoch}.npy"),
                    self.mask.color_filter.cpu().detach().numpy(),
                )

            torch.save(
                self.mask._optimizer.state_dict(), os.path.join(path, f"mask_optim_epoch{epoch}.pt")
            )

            psf_np = self.mask.get_psf().detach().cpu().numpy()[0, ...]
            psf_np = psf_np.squeeze()  # remove (potential) singleton color channel
            np.save(os.path.join(path, f"psf_epoch{epoch}.npy"), psf_np)
            save_image(psf_np, os.path.join(path, f"psf_epoch{epoch}.png"))
            plot_image(psf_np, gamma=self.gamma)
            plt.savefig(os.path.join(path, f"psf_epoch{epoch}_plot.png"))

        # save optimizer
        if include_optimizer:
            torch.save(self.optimizer.state_dict(), os.path.join(path, f"optim_epoch{epoch}.pt"))
        # save recon
        torch.save(self.recon.state_dict(), os.path.join(path, f"recon_epoch{epoch}"))



from datasets import load_dataset
import numpy as np
import requests
from PIL import Image
from io import BytesIO
import torch
from lensless.utils.image import resize
from lensless.utils.io import save_image
from huggingface_hub import hf_hub_download
from lensless.utils.io import load_psf
from torchvision import transforms


def get_image_url(file_name, repo_path):
    return (
        f"https://huggingface.co/datasets/{repo_path}/resolve/main/"
        + file_name
        + "?download=true"
    )

def load_url_image(url=None,):
    r"""

    Load an image from a URL and return a torch.Tensor.

    :param str url: URL of the image file.
    :param int, tuple[int] img_size: Size of the image to return.
    :param float downsample: Factor to downsample the image on each axis.
    :param bool grayscale: Whether to convert the image to grayscale.
    :param str resize_mode: If ``img_size`` is not None, options are ``"crop"`` or ``"resize"``.
    :param str device: Device on which to load the image (gpu or cpu).
    :return: :class:`torch.Tensor` containing the image.
    """

    response = requests.get(url)
    img = Image.open(BytesIO(response.content))
    x = np.array(img)
    return x



repo_id = "bezzam/DiffuserCam-Lensless-Mirflickr-Dataset"
downsample = 2

# load one example
dataset = load_dataset(repo_id, split="test")
h = dataset[0]["lensless"]
lensless_np = np.array(h)

# load PSF
# psf_url = get_image_url("psf.png", repo_id)
# # psf = load_url_image(psf_url)
# psf = psf / np.linalg.norm(psf.ravel())
# if downsample != 1:
#     psf = resize(psf, factor=1 / (downsample * 4))
# # add extra dimension to PSF
# psf = psf[np.newaxis, :]
# background subtraction
# bg = []
# bg_pix = (5, 25)
# for i in range(psf.shape[3]):
#     bg_i = np.mean(psf[:, bg_pix[0] : bg_pix[1], bg_pix[0] : bg_pix[1], i])
#     psf[:, :, :, i] -= bg_i
#     bg.append(bg_i)

# psf = np.clip(psf, a_min=0, a_max=psf.max())
# bg = np.array(bg)
# psf = torch.from_numpy(psf)

psf_fp = hf_hub_download(repo_id=repo_id, filename="psf.png", repo_type="dataset")
# psf_fp = hf_hub_download(repo_id=repo_id, filename="psf.tiff", repo_type="dataset")

psf, bg = load_psf(
    psf_fp,
    downsample=downsample * 4,  # PSF is 4x the resolution of the images
    return_float=True,
    return_bg=True,
    bg_pix=(0, 15),
)
transform_BRG2RGB = transforms.Lambda(lambda x: x[..., [2, 1, 0]])
psf = torch.from_numpy(psf)
# psf = transform_BRG2RGB(psf)



# prepare data
img = lensless_np / lensless_np.max()
if downsample != 1:
    img = resize(img, factor=1 / downsample)

img = img - bg
img = np.clip(img, a_min=0, a_max=img.max())
img = torch.from_numpy(img)
    
# perform reconstruction
from lensless import ADMM

recon = ADMM(psf, n_iter=100)
recon.set_data(img)
res = recon.apply(plot=False)

# save result
res = res.cpu().numpy()
save_image(res[0], "reconstruction_hf.png")

# ## same reconstruction with original dataset
# from lensless.utils.io import load_psf
# import torch
# from torchvision import transforms

# lensless_fp = "/scratch/bezzam/DiffuserCam_mirflickr/dataset/diffuser_images/im2.npy"
# psf_fp = "/home/bezzam/LenslessPiCam/data/psf/diffusercam_psf.tiff"
# psf, background = load_psf(
#     psf_fp,
#     downsample=downsample * 4,  # PSF is 4x the resolution of the images
#     return_float=True,
#     return_bg=True,
#     bg_pix=(0, 15),
# )
# transform_BRG2RGB = transforms.Lambda(lambda x: x[..., [2, 1, 0]])
# psf = transform_BRG2RGB(torch.from_numpy(psf))

# lensless = np.load(lensless_fp)
# if downsample != 1.0:
#     lensless = resize(lensless, factor=1 / downsample)
# lensless = torch.from_numpy(lensless)
# lensless = lensless.unsqueeze(0)
# lensless = transform_BRG2RGB(lensless)

# from lensless import ADMM
# recon = ADMM(psf, n_iter=100)
# recon.set_data(lensless)
# res = recon.apply(plot=False)

# # save result
# res = res.cpu().numpy()
# save_image(res[0], "reconstruction_original.png")


# import pudb; pudb.set_trace()


"""
To be run on the Raspberry Pi!
```
python scripts/measure/collect_dataset_on_device.py
```

Note that the script is configured for the  Raspberry Pi HQ camera

Parameters set in: configs/collect_dataset.yaml

To test on local machine, set dummy=True (which will just copy the files over).

"""

import numpy as np
import hydra
import time
import os
import pathlib as plib
import shutil
import tqdm
from picamerax import PiCamera
from fractions import Fraction
from PIL import Image
from lensless.utils.io import save_image

from lensless.hardware.constants import (
    RPI_HQ_CAMERA_CCM_MATRIX,
    RPI_HQ_CAMERA_BLACK_LEVEL,
)
import picamerax.array
from lensless.utils.image import bayer2rgb_cc, resize
import cv2


@hydra.main(version_base=None, config_path="../../configs", config_name="collect_dataset")
def collect_dataset(config):

    input_dir = config.input_dir
    output_dir = config.output_dir
    if output_dir is None:
        # create in same directory as input with timestamp
        output_dir = input_dir + "_measured_" + time.strftime("%Y%m%d-%H%M%S")

    MAX_TRIES = config.max_tries
    MIN_LEVEL = config.min_level
    MAX_LEVEL = config.max_level

    # if output dir exists check how many files done
    print(f"Output directory : {output_dir}")
    start_idx = 0
    if os.path.exists(output_dir):
        files = list(plib.Path(output_dir).glob(f"*.{config.output_file_ext}"))
        start_idx = len(files)
        print("\nNumber of completed measurements :", start_idx)

    # make output directory if need be
    output_dir = plib.Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    # assert input directory exists
    assert os.path.exists(input_dir)

    # get number of files with glob
    files = list(plib.Path(input_dir).glob(f"*.{config.input_file_ext}"))
    n_files = len(files)
    print(f"\nNumber of {config.input_file_ext} files :", n_files)
    if config.n_files:
        print(f"TEST : collecting first {config.n_files} files!")
        files = files[: config.n_files]
        n_files = len(files)

    if config.runtime:
        # convert to minutes
        runtime_min = config.runtime * 60
        runtime_sec = runtime_min * 60
        if config.runtime:
            print(f"\nScript will run for (at most) {config.runtime} hour(s).")

    if config.start_delay:
        # wait for this time before starting script
        delay = config.start_delay * 60
        start_time = time.time() + delay
        start_time = time.strftime("%H:%M:%S", time.localtime(start_time))
        print(f"\nScript will start at {start_time}")
        time.sleep(delay)

    print("\nStarting measurement!\n")
    start_time = time.time()

    if not config.dummy:

        res = config.capture.res
        down = config.capture.down

        # set up camera for consistent photos
        # https://picamera.readthedocs.io/en/release-1.13/recipes1.html#capturing-consistent-images
        # https://picamerax.readthedocs.io/en/latest/fov.html?highlight=camera%20resolution#sensor-modes
        # -- just get max resolution of camera
        camera = PiCamera(framerate=30)
        if res:
            assert len(res) == 2
        else:
            res = np.array(camera.MAX_RESOLUTION)
            if down is not None:
                res = (np.array(res) / down).astype(int)
        camera.close()

        # -- now set up camera with desired settings
        camera = PiCamera(
            framerate=1 / config.capture.exposure, sensor_mode=0, resolution=tuple(res)
        )

        # Set ISO to the desired value
        camera.resolution = tuple(res)
        camera.iso = config.capture.iso
        # Wait for the automatic gain control to settle
        time.sleep(config.capture.config_pause)
        # Now fix the values

        if config.capture.exposure:
            # in microseconds
            init_shutter_speed = int(config.capture.exposure * 1e6)
        else:
            init_shutter_speed = camera.exposure_speed
        camera.shutter_speed = init_shutter_speed
        camera.exposure_mode = "off"

        # AWB
        if config.capture.awb_gains:
            assert len(config.capture.awb_gains) == 2
            g = (Fraction(config.capture.awb_gains[0]), Fraction(config.capture.awb_gains[1]))
            g = tuple(g)
        else:
            g = camera.awb_gains

        camera.awb_mode = "off"
        camera.awb_gains = g

        # for parameters to settle
        time.sleep(5)

        print("Capturing at resolution: ", res)
        print("AWB gains", float(camera.awb_gains[0]), float(camera.awb_gains[1]))

    init_brightness = config.display.brightness

    # loop over files with tqdm
    exposure_vals = []
    brightness_vals = []
    n_tries_vals = []
    for i, _file in enumerate(tqdm.tqdm(files), start=start_idx):

        # save file in output directory as PNG
        output_fp = output_dir / _file.name
        output_fp = output_fp.with_suffix(f".{config.output_file_ext}")

        # if not done, perform measurement
        if not os.path.isfile(output_fp):

            if config.dummy:
                shutil.copyfile(_file, output_fp)
                time.sleep(1)

            else:

                # -- show on display
                screen_res = np.array(config.display.screen_res)
                hshift = config.display.hshift
                vshift = config.display.vshift
                pad = config.display.pad
                brightness = init_brightness
                display_image_path = config.display.output_fp
                rot90 = config.display.rot90
                os.system(
                    f"python scripts/measure/prep_display_image.py --fp {_file} --output_path {display_image_path} --screen_res {screen_res[0]} {screen_res[1]} --hshift {hshift} --vshift {vshift} --pad {pad} --brightness {brightness} --rot90 {rot90}"
                )

                time.sleep(config.capture.delay)

                # -- take picture
                max_pixel_val = 0
                fact = 2
                n_tries = 0

                camera.shutter_speed = init_shutter_speed
                time.sleep(5)

                current_screen_brightness = init_brightness
                current_shutter_speed = camera.shutter_speed
                print(f"current shutter speed: {current_shutter_speed}")
                print(f"current screen brightness: {current_screen_brightness}")

                while max_pixel_val < MIN_LEVEL or max_pixel_val > MAX_LEVEL:

                    if n_tries > MAX_TRIES:
                        print("Max number of tries reached!")
                        break

                    # get bayer data
                    stream = picamerax.array.PiBayerArray(camera)
                    camera.capture(stream, "jpeg", bayer=True)
                    output_bayer = np.sum(stream.array, axis=2).astype(np.uint16)

                    # convert to RGB
                    output = bayer2rgb_cc(
                        output_bayer,
                        nbits=12,
                        blue_gain=float(g[1]),
                        red_gain=float(g[0]),
                        black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
                        ccm=RPI_HQ_CAMERA_CCM_MATRIX,
                        nbits_out=8,
                    )

                    if down:
                        output = resize(
                            output[None, ...], factor=1 / down, interpolation=cv2.INTER_CUBIC
                        )[0]

                    # print range
                    print(f"{output_fp}, range: {output.min()} - {output.max()}")
                    max_pixel_val = output.max()

                    if max_pixel_val < MIN_LEVEL:
                        current_shutter_speed = init_shutter_speed * fact
                        camera.shutter_speed = current_shutter_speed
                        time.sleep(5)
                        print(f"increasing shutter speed to {current_shutter_speed}")
                        fact += 1

                    elif max_pixel_val > MAX_LEVEL:

                        current_screen_brightness = current_screen_brightness - 10

                        screen_res = np.array(config.display.screen_res)
                        hshift = config.display.hshift
                        vshift = config.display.vshift
                        pad = config.display.pad
                        brightness = current_screen_brightness
                        display_image_path = config.display.output_fp
                        rot90 = config.display.rot90
                        os.system(
                            f"python scripts/measure/prep_display_image.py --fp {_file} --output_path {display_image_path} --screen_res {screen_res[0]} {screen_res[1]} --hshift {hshift} --vshift {vshift} --pad {pad} --brightness {brightness} --rot90 {rot90}"
                        )
                        print(f"decreasing screen brightness to {current_screen_brightness}")

                        time.sleep(config.capture.delay)

                    n_tries += 1

                    # save image
                    save_image(output, output_fp)

                exposure_vals.append(current_shutter_speed / 1e6)
                brightness_vals.append(current_screen_brightness)
                n_tries_vals.append(n_tries)

        # check if runtime is exceeded
        if config.runtime:
            proc_time = time.time() - start_time
            if proc_time > runtime_sec:
                print(f"-- measured {i+1} / {n_files} files")
                break

    proc_time = time.time() - start_time
    print(f"\nFinished, {proc_time/60.:.3f} minutes.")

    # print brightness and exposure range and average
    print(f"brightness range: {np.min(brightness_vals)} - {np.max(brightness_vals)}")
    print(f"exposure range: {np.min(exposure_vals)} - {np.max(exposure_vals)}")
    print(f"n_tries range: {np.min(n_tries_vals)} - {np.max(n_tries_vals)}")
    print(f"brightness average: {np.mean(brightness_vals)}")
    print(f"exposure average: {np.mean(exposure_vals)}")
    print(f"n_tries average: {np.mean(n_tries_vals)}")


if __name__ == "__main__":
    collect_dataset()



"""
Apply ADMM reconstruction to folder.

```
python scripts/recon/dataset.py
```

To run APGD, use the following command:
```
python scripts/recon/dataset.py algo=apgd
```

To just copy resized raw data, use the following command:
```
python scripts/recon/dataset.py algo=null preprocess.data_dim=[48,64]
```

"""

import hydra
from hydra.utils import to_absolute_path
import os
import time
import numpy as np
from lensless.utils.io import load_psf, load_image, save_image
from lensless import ADMM
import torch
import glob
from tqdm import tqdm
from lensless.recon.apgd import APGD
from joblib import Parallel, delayed


@hydra.main(version_base=None, config_path="../../configs", config_name="recon_dataset")
def admm_dataset(config):

    algo = config.algo

    # get raw data file paths
    dataset = to_absolute_path(config.input.raw_data)
    if not os.path.isdir(dataset):
        print(f"No dataset found at {dataset}")
        try:
            from torchvision.datasets.utils import download_and_extract_archive
        except ImportError:
            exit()
        msg = "Do you want to download the sample CelebA dataset measured with a random Adafruit LCD pattern (1.2 GB)?"

        # default to yes if no input is given
        valid = input("%s (Y/n) " % msg).lower() != "n"
        if valid:
            url = "https://drive.switch.ch/index.php/s/m89D1tFEfktQueS/download"
            filename = "celeba_adafruit_random_2mm_20230720_1K.zip"
            download_and_extract_archive(
                url, os.path.dirname(dataset), filename=filename, remove_finished=True
            )
    data_fps = sorted(glob.glob(os.path.join(dataset, "*.png")))
    if config.n_files is not None:
        data_fps = data_fps[: config.n_files]
    n_files = len(data_fps)

    # load PSF
    psf_fp = to_absolute_path(config.input.psf)
    flip = config.preprocess.flip
    dtype = config.input.dtype
    print("\nPSF:")
    psf, bg = load_psf(
        psf_fp,
        verbose=True,
        downsample=config.preprocess.downsample,
        return_bg=True,
        flip=flip,
        dtype=dtype,
    )
    print(f"Downsampled PSF shape: {psf.shape}")

    data_dim = None
    if config.preprocess.data_dim is not None:
        data_dim = tuple(config.preprocess.data_dim) + (psf.shape[-1],)
    else:
        data_dim = psf.shape

    # -- create output folder
    output_folder = to_absolute_path(config.output_folder)
    if algo == "apgd":
        output_folder = output_folder + f"_apgd{config.apgd.max_iter}"
    elif algo == "admm":
        output_folder = output_folder + f"_admm{config.admm.n_iter}"
    else:
        output_folder = output_folder + "_raw"
    output_folder = output_folder + f"_{data_dim[-3]}x{data_dim[-2]}"
    os.makedirs(output_folder, exist_ok=True)

    # -- apply reconstruction
    if algo == "apgd":

        start_time = time.time()

        def recover(i):

            # reconstruction object
            recon = APGD(psf=psf, **config.apgd)

            data_fp = data_fps[i]

            # load data
            data = load_image(
                data_fp, flip=flip, bg=bg, as_4d=True, return_float=True, shape=data_dim
            )
            data = data[0]  # first depth

            # apply reconstruction
            recon.set_data(data)
            img = recon.apply(
                disp_iter=config.display.disp,
                gamma=config.display.gamma,
                plot=config.display.plot,
            )

            # -- extract region of interest and save
            if config.roi is not None:
                roi = config.roi
                img = img[roi[0] : roi[2], roi[1] : roi[3]]

            bn = os.path.basename(data_fp)
            output_fp = os.path.join(output_folder, bn)
            save_image(img, output_fp)

        n_jobs = config.apgd.n_jobs
        if n_jobs > 1:
            Parallel(n_jobs=n_jobs)(delayed(recover)(i) for i in range(n_files))
        else:
            for i in tqdm(range(n_files)):
                recover(i)

    else:

        if config.torch:
            torch_dtype = torch.float32
            torch_device = config.torch_device
            psf = torch.from_numpy(psf).type(torch_dtype).to(torch_device)

        # create reconstruction object
        recon = None
        if config.algo == "admm":
            recon = ADMM(psf, **config.admm)

        # loop over files and apply reconstruction
        start_time = time.time()

        for i in tqdm(range(n_files)):
            data_fp = data_fps[i]

            # load data
            data = load_image(
                data_fp, flip=flip, bg=bg, as_4d=True, return_float=True, shape=data_dim
            )

            if config.torch:
                data = torch.from_numpy(data).type(torch_dtype).to(torch_device)

            if recon is not None:

                # set data
                recon.set_data(data)

                # apply reconstruction
                res = recon.apply(
                    n_iter=config.admm.n_iter,
                    disp_iter=config.display.disp,
                    gamma=config.display.gamma,
                    plot=config.display.plot,
                )

            else:

                # copy resized raw data
                res = data

            # save reconstruction as PNG
            # -- take first depth
            if config.torch:
                img = res[0].cpu().numpy()
            else:
                img = res[0]

            # -- extract region of interest
            if config.roi is not None:
                img = img[config.roi[0] : config.roi[2], config.roi[1] : config.roi[3]]

            bn = os.path.basename(data_fp)
            output_fp = os.path.join(output_folder, bn)
            save_image(img, output_fp)

        print(f"Processing time : {time.time() - start_time} s")
        # time per file
        print(f"Time per file : {(time.time() - start_time) / n_files} s")
        print("Files saved to: ", output_folder)


if __name__ == "__main__":
    admm_dataset()
